# -*- coding: utf-8 -*-


import jax

from brainpy import errors
from brainpy.base.base import Base
from brainpy.base.collector import ArrayCollector
from brainpy.math.jax.jaxarray import JaxArray, TrainVar

__all__ = [
  'grad', 'value_and_grad',
  'Grad', 'ValueAndGrad',
]


def grad(func, vars=None, argnums=None, has_aux=None,
         holomorphic=False, allow_int=False, reduce_axes=()):
  """Automatic Gradient Computation in JAX backend.

  Creates a function which evaluates the gradient of ``fun``.

  Parameters
  ----------
  func : function, Base
    Function to be differentiated. Its arguments at positions specified by
    ``argnums`` should be arrays, scalars, or standard Python containers.
    Argument arrays in the positions specified by ``argnums`` must be of
    inexact (i.e., floating-point or complex) type. It should return a scalar
    (which includes arrays with shape ``()`` but not arrays with shape ``(1,)`` etc.)
  argnums : optional, integer or sequence of integers
    Specifies which positional argument(s) to differentiate with respect to (default 0).
  has_aux: optional, bool.
    Indicates whether ``fun`` returns a pair where the
    first element is considered the output of the mathematical function to be
    differentiated and the second element is auxiliary data. Default False.
  holomorphic: optional, bool.
    Indicates whether ``fun`` is promised to be
    holomorphic. If True, inputs and outputs must be complex. Default False.
  allow_int: optional, bool.
    Whether to allow differentiating with
    respect to integer valued inputs. The gradient of an integer input will
    have a trivial vector-space dtype (float0). Default False.
  reduce_axes: optional, tuple of axis names.
    If an axis is listed here, and
    ``fun`` implicitly broadcasts a value over that axis, the backward pass
    will perform a ``psum`` of the corresponding gradient. Otherwise, the
    gradient will be per-example over named axes. For example, if ``'batch'``
    is a named batch axis, ``grad(f, reduce_axes=('batch',))`` will create a
    function that computes the total gradient while ``grad(f)`` will create
    one that computes the per-example gradient.

  Returns
  -------
  func : function
    A function with the same arguments as ``fun``, that evaluates the gradient
    of ``fun``. If ``argnums`` is an integer then the gradient has the same
    shape and type as the positional argument indicated by that integer. If
    argnums is a tuple of integers, the gradient is a tuple of values with the
    same shapes and types as the corresponding arguments. If ``has_aux`` is True
    then a pair of (gradient, auxiliary_data) is returned.

  Examples
  --------

  >>> import brainpy as bp
  >>> grad_tanh = grad(bp.math.tanh)
  >>> print(grad_tanh(0.2))
  0.961043
  """
  # vars
  if vars is None:
    if isinstance(func, Base):
      vars = func.vars()

  # function
  if not callable(func):
    raise ValueError('Must be a callable object.')

  # gradient
  if vars is None:
    has_aux = False if has_aux is None else has_aux
    argnums = 0 if argnums is None else argnums
    return jax.grad(fun=func,
                    argnums=argnums,
                    has_aux=has_aux,
                    holomorphic=holomorphic,
                    allow_int=allow_int,
                    reduce_axes=reduce_axes)
  else:
    return Grad(fun=func,
                vars=vars,
                argnums=argnums,
                has_aux=has_aux,
                holomorphic=holomorphic,
                allow_int=allow_int,
                reduce_axes=reduce_axes)


def value_and_grad(func, vars=None, argnums=None, has_aux=None,
                   holomorphic=False, allow_int=False, reduce_axes=()):
  """Automatic Gradient Computation in JAX backend.

  Create a function which evaluates both ``fun`` and the gradient of ``fun``.

  Parameters
  ----------
  func : function, Base
    Function to be differentiated. Its arguments at positions specified by
    ``argnums`` should be arrays, scalars, or standard Python containers. It
    should return a scalar (which includes arrays with shape ``()`` but not
    arrays with shape ``(1,)`` etc.)
  argnums: optional, integer or sequence of integers
    Specifies which
    positional argument(s) to differentiate with respect to (default 0).
  has_aux: optional, bool
    Indicates whether ``fun`` returns a pair where the
    first element is considered the output of the mathematical function to be
    differentiated and the second element is auxiliary data. Default False.
  holomorphic: optional, bool
    Indicates whether ``fun`` is promised to be
    holomorphic. If True, inputs and outputs must be complex. Default False.
  allow_int: optional, bool
    Whether to allow differentiating with
    respect to integer valued inputs. The gradient of an integer input will
    have a trivial vector-space dtype (float0). Default False.
  reduce_axes: optional, tuple of axis names
    If an axis is listed here, and
    ``fun`` implicitly broadcasts a value over that axis, the backward pass
    will perform a ``psum`` of the corresponding gradient. Otherwise, the
    gradient will be per-example over named axes. For example, if ``'batch'``
    is a named batch axis, ``value_and_grad(f, reduce_axes=('batch',))`` will
    create a function that computes the total gradient while
    ``value_and_grad(f)`` will create one that computes the per-example
    gradient.

  Returns
  -------
    A function with the same arguments as ``fun`` that evaluates both ``fun``
    and the gradient of ``fun`` and returns them as a pair (a two-element
    tuple). If ``argnums`` is an integer then the gradient has the same shape
    and type as the positional argument indicated by that integer. If argnums is
    a sequence of integers, the gradient is a tuple of values with the same
    shapes and types as the corresponding arguments.
  """
  # function
  if not callable(func):
    raise ValueError('Must be a callable object.')

  # vars
  if vars is None:
    if isinstance(func, Base):
      vars = func.vars()
    elif hasattr(func, '__self__') and isinstance(func.__self__, Base):
      vars = func.__self__.vars()

  # jit compilation
  if vars is None:
    has_aux = False if has_aux is None else has_aux
    argnums = 0 if argnums is None else argnums
    return jax.value_and_grad(fun=func,
                              argnums=argnums,
                              has_aux=has_aux,
                              holomorphic=holomorphic,
                              allow_int=allow_int,
                              reduce_axes=reduce_axes)
  else:
    return ValueAndGrad(fun=func,
                        vars=vars,
                        argnums=argnums,
                        has_aux=has_aux,
                        holomorphic=holomorphic,
                        allow_int=allow_int,
                        reduce_axes=reduce_axes)


class Gradient(Base):
  """Base Class to Compute Gradients."""

  def __init__(self, raw, vars=None, argnums=None, has_aux=None,
               holomorphic=False, allow_int=False, reduce_axes=()):
    super(Gradient, self).__init__()

    # 'raw'
    self.raw = raw

    # trainable variables and dynamical variables
    if vars is None:
      if not isinstance(raw, Base):
        raise errors.BrainPyError(f'When "vars" is not provided, "raw" must be a {Base} object.')
      vars = raw.vars().unique()
    _train_vars = ArrayCollector()
    _dyn_vars = ArrayCollector()
    for key, var in vars.items():
      if isinstance(var, TrainVar):
        _train_vars[key] = var
      else:
        _dyn_vars[key] = var
    self._train_vars = _train_vars.unique()
    self._dyn_vars = _dyn_vars.unique()

    # 'argnums'
    if argnums is None:
      argnums = (0,)
    elif isinstance(argnums, int):
      argnums = (0, argnums + 2)
    else:
      argnums = (0,) + tuple(a + 2 for a in argnums)
    self.argnums = argnums

    # others
    self.has_aux = False if has_aux is None else True
    self.holomorphic = holomorphic
    self.allow_int = allow_int
    self.reduce_axes = reduce_axes

    # signature
    # signature = inspect.signature(raw)
    # self.__signature__ = signature.replace(return_annotation=Tuple[List[JaxArray], signature.return_annotation])

    # final functions
    if self.has_aux:
      # Users should return the auxiliary data like:
      # ------------
      # >>> # 1. example of return one data
      # >>> return scalar_loss, data
      # >>> # 2. example of return multiple data
      # >>> return scalar_loss, (data1, data2, ...)
      def func(train_vars, dyn_vars, *args, **kwargs):
        self._train_vars.assign(train_vars)
        self._dyn_vars.assign(dyn_vars)
        # outputs: [0] is the value for gradient,
        #          [1] is other values for return
        outputs = self.raw(*args, **kwargs)
        output = outputs[0].value if isinstance(outputs[0], JaxArray) else outputs[0]
        return output, (outputs, self._train_vars.dict(), self._dyn_vars.dict())
    else:
      # Users should return the scalar value like this:
      # ------------
      # >>> return scalar_loss
      def func(train_vars, dyn_vars, *args, **kwargs):
        self._train_vars.assign(train_vars)
        self._dyn_vars.assign(dyn_vars)
        output = self.raw(*args, **kwargs)
        output2 = output.value if isinstance(output, JaxArray) else output
        return output2, (output, self._train_vars.dict(), self._dyn_vars.dict())

    # function for gradient
    self._call = jax.grad(fun=func,
                          argnums=self.argnums,
                          has_aux=True,
                          holomorphic=self.holomorphic,
                          allow_int=self.allow_int,
                          reduce_axes=self.reduce_axes)


class Grad(Gradient):
  """Compute the gradients of trainable variables for the given object.

  Examples
  --------

  This example is that we return two auxiliary data, i.e., ``has_aux=True``.

  >>> import brainpy as bp
  >>>
  >>> class Test(bp.dnn.Module):
  >>>   def __init__(self):
  >>>     super(Test, self).__init__()
  >>>     self.a = bp.TrainVar(bp.math.ones(1))
  >>>     self.b = bp.TrainVar(bp.math.ones(1))
  >>>
  >>>   def __call__(self, c):
  >>>     ab = self.a * self.b
  >>>     ab2 = ab * 2
  >>>     vv = ab2 + c
  >>>     return vv, (ab, ab2)
  >>>
  >>> test = Test()
  >>> test_grad = Grad(test, test.vars(), argnums=0, has_aux=True)
  >>> grads, outputs = test_grad(10.)
  >>> grads
  (DeviceArray(1., dtype=float32),
   {'Test3.a': DeviceArray([2.], dtype=float32), 'Test3.b': DeviceArray([2.], dtype=float32)})
  >>> outputs
  (JaxArray(DeviceArray([1.], dtype=float32)),
   JaxArray(DeviceArray([2.], dtype=float32)))
  """

  def __init__(self, fun, vars=None, argnums=None, has_aux=None,
               holomorphic=False, allow_int=False, reduce_axes=()):
    super(Grad, self).__init__(raw=fun,
                               vars=vars,
                               argnums=argnums,
                               has_aux=has_aux,
                               holomorphic=holomorphic,
                               allow_int=allow_int,
                               reduce_axes=reduce_axes)

  def __call__(self, *args, **kwargs):
    grads, (outputs, train_vars, dyn_vars) = self._call(self._train_vars.dict(),
                                                        self._dyn_vars.dict(),
                                                        *args,
                                                        **kwargs)
    self._train_vars.assign(train_vars)
    self._dyn_vars.assign(dyn_vars)
    grads = grads[0] if len(self.argnums) == 1 else grads[1:] + grads[:1]
    return (grads, outputs[1]) if self.has_aux else grads


class ValueAndGrad(Gradient):
  """Compute the results and the gradients of trainable variables for the given object.

  Examples
  --------

  This example is that we return two auxiliary data, i.e., ``has_aux=True``.

  >>> import brainpy as bp
  >>>
  >>> class Test(bp.dnn.Module):
  >>>   def __init__(self):
  >>>     super(Test, self).__init__()
  >>>     self.a = bp.TrainVar(bp.math.ones(1))
  >>>     self.b = bp.TrainVar(bp.math.ones(1))
  >>>
  >>>   def __call__(self, c):
  >>>     ab = self.a * self.b
  >>>     ab2 = ab * 2
  >>>     vv = ab2 + c
  >>>     return vv, (ab, ab2)
  >>>
  >>> test = Test()
  >>> test_grad = ValueAndGrad(test, argnums=0, has_aux=True)
  >>> outputs, grads = test_grad(10.)
  >>> grads
  (DeviceArray(1., dtype=float32),
   {'Test3.a': DeviceArray([2.], dtype=float32), 'Test3.b': DeviceArray([2.], dtype=float32)})
  >>> outputs
  (JaxArray(DeviceArray(12., dtype=float32)),
   (JaxArray(DeviceArray([1.], dtype=float32)),
    JaxArray(DeviceArray([2.], dtype=float32))))
  """

  def __init__(self, fun, vars=None, argnums=None, has_aux=None, holomorphic=False,
               allow_int=False, reduce_axes=()):
    super(ValueAndGrad, self).__init__(raw=fun,
                                       vars=vars,
                                       argnums=argnums,
                                       has_aux=has_aux,
                                       holomorphic=holomorphic,
                                       allow_int=allow_int,
                                       reduce_axes=reduce_axes)

  def __call__(self, *args, **kwargs):
    grads, (outputs, train_vars, dyn_vars) = self._call(self._train_vars.dict(),
                                                        self._dyn_vars.dict(),
                                                        *args,
                                                        **kwargs)
    self._train_vars.assign(train_vars)
    self._dyn_vars.assign(dyn_vars)
    grads = grads[0] if len(self.argnums) == 1 else grads[1:] + grads[:1]
    return outputs, grads



# -*- coding: utf-8 -*-

from . import fft
from . import linalg
from . import random

from .code import *
from .compilation import *
from .controls import *
from .function import *
from .gradient import *
from .ndarray import *
from .ops import *

try:
  from . import overload
  from . import ast2numba
except ModuleNotFoundError:
  pass



# -*- coding: utf-8 -*-

"""
Numerical methods for ordinary differential equations.
"""

from . import adaptive_rk
from . import explicit_rk
from . import exponential
from .adaptive_rk import *
from .explicit_rk import *
from .exponential import *



# -*- coding: utf-8 -*-

from brainpy import errors, math
from brainpy.integrators import constants
from . import common

__all__ = [
  'srk1w1_scalar',
  'srk2w1_scalar',
  'KlPl_scalar',
]


class Tools(object):
  @staticmethod
  def _noise_terms(code_lines, variables, vdt, triple_integral=True):
    # num_vars = len(variables)
    # if num_vars > 1:
    #     code_lines.append(f'  all_I1 = math.normal(0.0, dt_sqrt, ({num_vars},)+math.shape({variables[0]}))')
    #     code_lines.append(f'  all_I0 = math.normal(0.0, dt_sqrt, ({num_vars},)+math.shape({variables[0]}))')
    #     code_lines.append(f'  all_I10 = 0.5 * {vdt} * (all_I1 + all_I0 / 3.0 ** 0.5)')
    #     code_lines.append(f'  all_I11 = 0.5 * (all_I1 ** 2 - {vdt})')
    #     if triple_integral:
    #         code_lines.append(f'  all_I111 = (all_I1 ** 3 - 3 * {vdt} * all_I1) / 6')
    #     code_lines.append(f'  ')
    #     for i, var in enumerate(variables):
    #         code_lines.append(f'  {var}_I1 = all_I1[{i}]')
    #         code_lines.append(f'  {var}_I0 = all_I0[{i}]')
    #         code_lines.append(f'  {var}_I10 = all_I10[{i}]')
    #         code_lines.append(f'  {var}_I11 = all_I11[{i}]')
    #         if triple_integral:
    #             code_lines.append(f'  {var}_I111 = all_I111[{i}]')
    #         code_lines.append(f'  ')
    # else:
    #     var = variables[0]
    #     code_lines.append(f'  {var}_I1 = math.normal(0.0, dt_sqrt, math.shape({var}))')
    #     code_lines.append(f'  {var}_I0 = math.normal(0.0, dt_sqrt, math.shape({var}))')
    #     code_lines.append(f'  {var}_I10 = 0.5 * {vdt} * ({var}_I1 + {var}_I0 / 3.0 ** 0.5)')
    #     code_lines.append(f'  {var}_I11 = 0.5 * ({var}_I1 ** 2 - {vdt})')
    #     if triple_integral:
    #         code_lines.append(f'  {var}_I111 = ({var}_I1 ** 3 - 3 * {vdt} * {var}_I1) / 6')
    #     code_lines.append('  ')

    for var in variables:
      code_lines.append(f'  {var}_I1 = math.random.normal(0.000, dt_sqrt, math.shape({var}))')
      code_lines.append(f'  {var}_I0 = math.random.normal(0.000, dt_sqrt, math.shape({var}))')
      code_lines.append(f'  {var}_I10 = 0.5 * {vdt} * ({var}_I1 + {var}_I0 / 3.0 ** 0.5)')
      code_lines.append(f'  {var}_I11 = 0.5 * ({var}_I1 ** 2 - {vdt})')
      if triple_integral:
        code_lines.append(f'  {var}_I111 = ({var}_I1 ** 3 - 3 * {vdt} * {var}_I1) / 6')
      code_lines.append('  ')

  @staticmethod
  def _state1(code_lines, variables, parameters):
    f_names = [f'{var}_f_H0s1' for var in variables]
    g_names = [f'{var}_g_H1s1' for var in variables]
    code_lines.append(f'  {", ".join(f_names)} = f({", ".join(variables + parameters)})')
    code_lines.append(f'  {", ".join(g_names)} = g({", ".join(variables + parameters)})')
    code_lines.append('  ')


class Wrappers(object):
  @staticmethod
  def srk1w1(f, g, dt, show_code, intg_type, var_type, wiener_type):
    vdt, variables, parameters, arguments, func_name = common.basic_info(f=f, g=g)
    arguments = list(arguments) + [f'{vdt}={dt}']

    # 1. code scope
    code_scope = {'f': f, 'g': g, 'math': math}

    # 2. code lines
    code_lines = [f'def {func_name}({", ".join(arguments)}):']
    code_lines.append(f'  {vdt}_sqrt = {vdt} ** 0.5')

    # 2.1 noise
    Tools._noise_terms(code_lines, variables, vdt, triple_integral=True)

    # 2.2 stage 1
    Tools._state1(code_lines, variables, parameters)

    # 2.3 stage 2
    all_H0s2, all_H1s2 = [], []
    for var in variables:
      code_lines.append(f'  {var}_H0s2 = {var} + {vdt} * 0.75 * {var}_f_H0s1 + '
                        f'1.5 * {var}_g_H1s1 * {var}_I10 / {vdt}')
      all_H0s2.append(f'{var}_H0s2')
      code_lines.append(f'  {var}_H1s2 = {var} + {vdt} * 0.25 * {var}_f_H0s1 + '
                        f'dt_sqrt * 0.5 * {var}_g_H1s1')
      all_H1s2.append(f'{var}_H1s2')
    all_H0s2.append(f't + 0.75 * {vdt}')  # t
    all_H1s2.append(f't + 0.25 * {vdt}')  # t
    f_names = [f'{var}_f_H0s2' for var in variables]
    code_lines.append(f'  {", ".join(f_names)} = f({", ".join(all_H0s2 + parameters[1:])})')
    g_names = [f'{var}_g_H1s2' for var in variables]
    code_lines.append(f'  {", ".join(g_names)} = g({", ".join(all_H1s2 + parameters[1:])})')
    code_lines.append('  ')

    # 2.4 state 3
    all_H1s3 = []
    for var in variables:
      code_lines.append(f'  {var}_H1s3 = {var} + {vdt} * {var}_f_H0s1 - dt_sqrt * {var}_g_H1s1')
      all_H1s3.append(f'{var}_H1s3')
    all_H1s3.append(f't + {vdt}')  # t
    g_names = [f'{var}_g_H1s3' for var in variables]
    code_lines.append(f'  {", ".join(g_names)} = g({", ".join(all_H1s3 + parameters[1:])})')
    code_lines.append('  ')

    # 2.5 state 4
    all_H1s4 = []
    for var in variables:
      code_lines.append(f'  {var}_H1s4 = {var} + 0.25 * {vdt} * {var}_f_H0s1 + dt_sqrt * '
                        f'(-5 * {var}_g_H1s1 + 3 * {var}_g_H1s2 + 0.5 * {var}_g_H1s3)')
      all_H1s4.append(f'{var}_H1s4')
    all_H1s4.append(f't + 0.25 * {vdt}')  # t
    g_names = [f'{var}_g_H1s4' for var in variables]
    code_lines.append(f'  {", ".join(g_names)} = g({", ".join(all_H1s4 + parameters[1:])})')
    code_lines.append('  ')

    # 2.6 final stage
    for var in variables:
      code_lines.append(f'  {var}_f1 = {var}_f_H0s1/3 + {var}_f_H0s2 * 2/3')
      code_lines.append(
        f'  {var}_g1 = -{var}_I1 - {var}_I11/dt_sqrt + 2 * {var}_I10/{vdt} - 2 * {var}_I111/{vdt}')
      code_lines.append(f'  {var}_g2 = {var}_I1 * 4/3 + {var}_I11 / dt_sqrt * 4/3 - '
                        f'{var}_I10 / {vdt} * 4/3 + {var}_I111 / {vdt} * 5/3')
      code_lines.append(f'  {var}_g3 = {var}_I1 * 2/3 - {var}_I11/dt_sqrt/3 - '
                        f'{var}_I10 / {vdt} * 2/3 - {var}_I111 / {vdt} * 2/3')
      code_lines.append(f'  {var}_g4 = {var}_I111 / {vdt}')
      code_lines.append(f'  {var}_new = {var} + {vdt} * {var}_f1 + {var}_g1 * {var}_g_H1s1 + '
                        f'{var}_g2 * {var}_g_H1s2 + {var}_g3 * {var}_g_H1s3 + {var}_g4 * {var}_g_H1s4')
      code_lines.append('  ')

    # returns
    new_vars = [f'{var}_new' for var in variables]
    code_lines.append(f'  return {", ".join(new_vars)}')

    # return and compile
    return common.compile_and_assign_attrs(
      code_lines=code_lines, code_scope=code_scope, show_code=show_code,
      variables=variables, parameters=parameters, func_name=func_name,
      intg_type=intg_type, var_type=var_type, wiener_type=wiener_type,
      dt=dt, method='srk1w1', raw_func=dict(f=f, g=g))

  @staticmethod
  def srk2w1(f, g, dt, show_code, intg_type, var_type, wiener_type):
    vdt, variables, parameters, arguments, func_name = common.basic_info(f=f, g=g)
    arguments = list(arguments) + [f'{vdt}={dt}']

    # 1. code scope
    code_scope = {'f': f, 'g': g, 'math': math}

    # 2. code lines
    code_lines = [f'def {func_name}({", ".join(arguments)}):']
    code_lines.append(f'  {vdt}_sqrt = {vdt} ** 0.5')

    # 2.1 noise
    Tools._noise_terms(code_lines, variables, vdt, triple_integral=True)

    # 2.2 stage 1
    Tools._state1(code_lines, variables, parameters)

    # 2.3 stage 2
    # ----
    # H0s2 = x + dt * f_H0s1
    # H1s2 = x + dt * 0.25 * f_H0s1 - dt_sqrt * 0.5 * g_H1s1
    # f_H0s2 = f(H0s2, t + dt, *args)
    # g_H1s2 = g(H1s2, t + 0.25 * dt, *args)
    all_H0s2, all_H1s2 = [], []
    for var in variables:
      code_lines.append(f'  {var}_H0s2 = {var} + {vdt} * {var}_f_H0s1')
      all_H0s2.append(f'{var}_H0s2')
      code_lines.append(f'  {var}_H1s2 = {var} + {vdt} * 0.25 * {var}_f_H0s1 - '
                        f'dt_sqrt * 0.5 * {var}_g_H1s1')
      all_H1s2.append(f'{var}_H1s2')
    all_H0s2.append(f't + {vdt}')  # t
    all_H1s2.append(f't + 0.25 * {vdt}')  # t
    f_names = [f'{var}_f_H0s2' for var in variables]
    code_lines.append(f'  {", ".join(f_names)} = f({", ".join(all_H0s2 + parameters[1:])})')
    g_names = [f'{var}_g_H1s2' for var in variables]
    code_lines.append(f'  {", ".join(g_names)} = g({", ".join(all_H1s2 + parameters[1:])})')
    code_lines.append('  ')

    # 2.4 state 3
    # ---
    # H0s3 = x + dt * (0.25 * f_H0s1 + 0.25 * f_H0s2) + (g_H1s1 + 0.5 * g_H1s2) * I10 / dt
    # H1s3 = x + dt * f_H0s1 + dt_sqrt * g_H1s1
    # f_H0s3 = g(H0s3, t + 0.5 * dt, *args)
    # g_H1s3 = g(H1s3, t + dt, *args)
    all_H0s3, all_H1s3 = [], []
    for var in variables:
      code_lines.append(f'  {var}_H0s3 = {var} + {vdt} * (0.25 * {var}_f_H0s1 + 0.25 * {var}_f_H0s2) + '
                        f'({var}_g_H1s1 + 0.5 * {var}_g_H1s2) * {var}_I10 / {vdt}')
      all_H0s3.append(f'{var}_H0s3')
      code_lines.append(f'  {var}_H1s3 = {var} + {vdt} * {var}_f_H0s1 + dt_sqrt * {var}_g_H1s1')
      all_H1s3.append(f'{var}_H1s3')
    all_H0s3.append(f't + 0.5 * {vdt}')  # t
    all_H1s3.append(f't + {vdt}')  # t
    f_names = [f'{var}_f_H0s3' for var in variables]
    g_names = [f'{var}_g_H1s3' for var in variables]
    code_lines.append(f'  {", ".join(f_names)} = f({", ".join(all_H0s3 + parameters[1:])})')
    code_lines.append(f'  {", ".join(g_names)} = g({", ".join(all_H1s3 + parameters[1:])})')
    code_lines.append('  ')

    # 2.5 state 4
    # ----
    # H1s4 = x + dt * 0.25 * f_H0s3 + dt_sqrt * (2 * g_H1s1 - g_H1s2 + 0.5 * g_H1s3)
    # g_H1s4 = g(H1s4, t + 0.25 * dt, *args)
    all_H1s4 = []
    for var in variables:
      code_lines.append(f'  {var}_H1s4 = {var} + 0.25 * {vdt} * {var}_f_H0s1 + dt_sqrt * '
                        f'(2 * {var}_g_H1s1 - {var}_g_H1s2 + 0.5 * {var}_g_H1s3)')
      all_H1s4.append(f'{var}_H1s4')
    all_H1s4.append(f't + 0.25 * {vdt}')  # t
    g_names = [f'{var}_g_H1s4' for var in variables]
    code_lines.append(f'  {", ".join(g_names)} = g({", ".join(all_H1s4 + parameters[1:])})')
    code_lines.append('  ')

    # 2.6 final stage
    # ----
    # f1 = f_H0s1 / 6 + f_H0s2 / 6 + f_H0s3 * 2 / 3
    # g1 = - I1 + I11 / dt_sqrt + 2 * I10 / dt - 2 * I111 / dt
    # g2 = I1 * 4 / 3 - I11 / dt_sqrt * 4 / 3 - I10 / dt * 4 / 3 + I111 / dt * 5 / 3
    # g3 = I1 * 2 / 3 + I11 / dt_sqrt / 3 - I10 / dt * 2 / 3 - I111 / dt * 2 / 3
    # g4 = I111 / dt
    # y1 = x + dt * f1 + g1 * g_H1s1 + g2 * g_H1s2 + g3 * g_H1s3 + g4 * g_H1s4
    for var in variables:
      code_lines.append(f'  {var}_f1 = {var}_f_H0s1/6 + {var}_f_H0s2/6 + {var}_f_H0s3*2/3')
      code_lines.append(
        f'  {var}_g1 = -{var}_I1 + {var}_I11/dt_sqrt + 2 * {var}_I10/{vdt} - 2 * {var}_I111/{vdt}')
      code_lines.append(f'  {var}_g2 = {var}_I1 * 4/3 - {var}_I11 / dt_sqrt * 4/3 - '
                        f'{var}_I10 / {vdt} * 4/3 + {var}_I111 / {vdt} * 5/3')
      code_lines.append(f'  {var}_g3 = {var}_I1 * 2/3 + {var}_I11/dt_sqrt/3 - '
                        f'{var}_I10 / {vdt} * 2/3 - {var}_I111 / {vdt} * 2/3')
      code_lines.append(f'  {var}_g4 = {var}_I111 / {vdt}')
      code_lines.append(f'  {var}_new = {var} + {vdt} * {var}_f1 + {var}_g1 * {var}_g_H1s1 + '
                        f'{var}_g2 * {var}_g_H1s2 + {var}_g3 * {var}_g_H1s3 + {var}_g4 * {var}_g_H1s4')
      code_lines.append('  ')

    # returns
    new_vars = [f'{var}_new' for var in variables]
    code_lines.append(f'  return {", ".join(new_vars)}')

    # return and compile
    return common.compile_and_assign_attrs(
      code_lines=code_lines, code_scope=code_scope, show_code=show_code,
      variables=variables, parameters=parameters, func_name=func_name,
      intg_type=intg_type, var_type=var_type, wiener_type=wiener_type,
      dt=dt, method='srk2w1', raw_func=dict(f=f, g=g))

  @staticmethod
  def KlPl(f, g, dt, show_code, intg_type, var_type, wiener_type):
    vdt, variables, parameters, arguments, func_name = common.basic_info(f=f, g=g)
    arguments = list(arguments) + [f'{vdt}={dt}']

    # 1. code scope
    code_scope = {'f': f, 'g': g, 'math': math}

    # 2. code lines
    code_lines = [f'def {func_name}({", ".join(arguments)}):']
    code_lines.append(f'  {vdt}_sqrt = {vdt} ** 0.5')

    # 2.1 noise
    Tools._noise_terms(code_lines, variables, vdt, triple_integral=False)

    # 2.2 stage 1
    Tools._state1(code_lines, variables, parameters)

    # 2.3 stage 2
    # ----
    # H1s2 = x + dt * f_H0s1 + dt_sqrt * g_H1s1
    # g_H1s2 = g(H1s2, t0, *args)
    all_H1s2 = []
    for var in variables:
      code_lines.append(f'  {var}_H1s2 = {var} + {vdt} * {var}_f_H0s1 + dt_sqrt * {var}_g_H1s1')
      all_H1s2.append(f'{var}_H1s2')
    g_names = [f'{var}_g_H1s2' for var in variables]
    code_lines.append(f'  {", ".join(g_names)} = g({", ".join(all_H1s2 + parameters)})')
    code_lines.append('  ')

    # 2.4 final stage
    # ----
    # g1 = (I1 - I11 / dt_sqrt + I10 / dt)
    # g2 = I11 / dt_sqrt
    # y1 = x + dt * f_H0s1 + g1 * g_H1s1 + g2 * g_H1s2
    for var in variables:
      code_lines.append(f'  {var}_g1 = -{var}_I1 + {var}_I11/dt_sqrt + {var}_I10/{vdt}')
      code_lines.append(f'  {var}_g2 = {var}_I11 / dt_sqrt')
      code_lines.append(f'  {var}_new = {var} + {vdt} * {var}_f_H0s1 + '
                        f'{var}_g1 * {var}_g_H1s1 + {var}_g2 * {var}_g_H1s2')
      code_lines.append('  ')

    # returns
    new_vars = [f'{var}_new' for var in variables]
    code_lines.append(f'  return {", ".join(new_vars)}')

    # return and compile
    return common.compile_and_assign_attrs(
      code_lines=code_lines, code_scope=code_scope, show_code=show_code,
      variables=variables, parameters=parameters, func_name=func_name,
      intg_type=intg_type, var_type=var_type, wiener_type=wiener_type,
      dt=dt, method='KlPl', raw_func=dict(f=f, g=g))

  @staticmethod
  def wrap(wrapper, f, g, dt, intg_type, var_type, wiener_type, show_code):
    """The base function to format a SRK method.

    Parameters
    ----------
    f : callable
        The drift function of the SDE_INT.
    g : callable
        The diffusion function of the SDE_INT.
    dt : float
        The numerical precision.
    intg_type : str
        "utils.ITO_SDE" : Ito's Stochastic Calculus.
        "utils.STRA_SDE" : Stratonovich's Stochastic Calculus.
    wiener_type : str
    var_type : str
        "scalar" : with the shape of ().
        "population" : with the shape of (N,) or (N1, N2) or (N1, N2, ...).
        "system": with the shape of (d, ), (d, N), or (d, N1, N2).
    show_code : bool
        Whether show the formatted code.

    Returns
    -------
    numerical_func : callable
        The numerical function.
    """

    var_type = constants.POP_VAR if var_type is None else var_type
    intg_type = constants.ITO_SDE if intg_type is None else intg_type
    wiener_type = constants.SCALAR_WIENER if wiener_type is None else wiener_type
    if var_type not in constants.SUPPORTED_VAR_TYPE:
      raise errors.IntegratorError(f'Currently, BrainPy only supports variable types: '
                                   f'{constants.SUPPORTED_VAR_TYPE}. But we got {var_type}.')
    if intg_type != constants.ITO_SDE:
      raise errors.IntegratorError(f'SRK method for SDEs with scalar noise only supports Ito SDE_INT type, '
                                   f'but we got {intg_type} integral.')
    if wiener_type != constants.SCALAR_WIENER:
      raise errors.IntegratorError(f'SRK method for SDEs with scalar noise only supports scalar '
                                   f'Wiener Process, but we got "{wiener_type}" noise.')

    show_code = False if show_code is None else show_code
    dt = math.get_dt() if dt is None else dt

    if f is not None and g is not None:
      return wrapper(f=f, g=g, dt=dt, show_code=show_code, intg_type=intg_type,
                     var_type=var_type, wiener_type=wiener_type)

    elif f is not None:
      return lambda g: wrapper(f=f, g=g, dt=dt, show_code=show_code, intg_type=intg_type,
                               var_type=var_type, wiener_type=wiener_type)

    elif g is not None:
      return lambda f: wrapper(f=f, g=g, dt=dt, show_code=show_code, intg_type=intg_type,
                               var_type=var_type, wiener_type=wiener_type)

    else:
      raise ValueError('Must provide "f" or "g".')


def srk1w1_scalar(f=None, g=None, dt=None, intg_type=None, var_type=None, wiener_type=None, show_code=None):
  """Order 2.0 weak SRK methods for SDEs with scalar Wiener process.

  This method has have strong orders :backend:`(p_d, p_s) = (2.0,1.5)`.

  The Butcher table is:

  .. math::

      \\begin{array}{l|llll|llll|llll}
          0   &&&&&  &&&&  &&&& \\\\
          3/4 &3/4&&&& 3/2&&& &&&& \\\\
          0   &0&0&0&& 0&0&0&& &&&&\\\\
          \\hline
          0 \\\\
          1/4 & 1/4&&& & 1/2&&&\\\\
          1 & 1&0&&& -1&0&\\\\
          1/4& 0&0&1/4&&  -5&3&1/2\\\\
          \\hline
          & 1/3& 2/3& 0 & 0 & -1 & 4/3 & 2/3&0 & -1 &4/3 &-1/3 &0 \\\\
          \\hline
          & &&&& 2 &-4/3 & -2/3 & 0 & -2 & 5/3 & -2/3 & 1
      \\end{array}


  References
  ----------

  .. [1] Rößler, Andreas. "Strong and weak approximation methods for stochastic differential
          equations—some recent developments." Recent developments in applied probability and
          statistics. Physica-Verlag HD, 2010. 127-153.
  .. [2] Rößler, Andreas. "Runge–Kutta methods for the strong approximation of solutions of
          stochastic differential equations." SIAM Journal on Numerical Analysis 48.3
          (2010): 922-952.

  """
  return Wrappers.wrap(Wrappers.srk1w1, f=f, g=g, dt=dt, intg_type=intg_type, var_type=var_type,
                       wiener_type=wiener_type, show_code=show_code)


def srk2w1_scalar(f=None, g=None, dt=None, intg_type=None, var_type=None, wiener_type=None, show_code=None):
  """Order 1.5 Strong SRK Methods for SDEs witdt Scalar Noise.

  This method has have strong orders :backend:`(p_d, p_s) = (3.0,1.5)`.

  The Butcher table is:

  .. math::

      \\begin{array}{c|cccc|cccc|ccc|}
          0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & & & & \\\\
          1 & 1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & & & & \\\\
          1 / 2 & 1 / 4 & 1 / 4 & 0 & 0 & 1 & 1 / 2 & 0 & 0 & & & & \\\\
          0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & & & & \\\\
          \\hline 0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & & & & \\\\
          1 / 4 & 1 / 4 & 0 & 0 & 0 & -1 / 2 & 0 & 0 & 0 & & & & \\\\
          1 & 1 & 0 & 0 & 0 & 1 & 0 & 0 & 0 & & & & \\\\
          1 / 4 & 0 & 0 & 1 / 4 & 0 & 2 & -1 & 1 / 2 & 0 & & & & \\\\
          \\hline & 1 / 6 & 1 / 6 & 2 / 3 & 0 & -1 & 4 / 3 & 2 / 3 & 0 & -1 & -4 / 3 & 1 / 3 & 0 \\\\
          \\hline & & & & &2 & -4 / 3 & -2 / 3 & 0 & -2 & 5 / 3 & -2 / 3 & 1
      \\end{array}


  References
  ----------

  [1] Rößler, Andreas. "Strong and weak approximation methods for stochastic differential
      equations—some recent developments." Recent developments in applied probability and
      statistics. Physica-Verlag HD, 2010. 127-153.
  [2] Rößler, Andreas. "Runge–Kutta methods for the strong approximation of solutions of
      stochastic differential equations." SIAM Journal on Numerical Analysis 48.3
      (2010): 922-952.
  """
  return Wrappers.wrap(Wrappers.srk2w1, f=f, g=g, dt=dt, intg_type=intg_type, var_type=var_type,
                       wiener_type=wiener_type, show_code=show_code)


def KlPl_scalar(f=None, g=None, dt=None, intg_type=None, var_type=None, wiener_type=None, show_code=None):
  """Order 1.0 Strong SRK Methods for SDEs with Scalar Noise.

  This method has have orders :backend:`p_s = 1.0`.

  The Butcher table is:

  .. math::

      \\begin{array}{c|cc|cc|cc|c}
          0 & 0 & 0 & 0 & 0 & & \\\\
          0 & 0 & 0 & 0 & 0 & & \\\\
          \\hline 0 & 0 & 0 & 0 & 0 & & \\\\
          0 & 1 & 0 & 1 & 0 & & \\\\
          \\hline 0 & 1 & 0 & 1 & 0 & -1 & 1 \\\\
          \\hline & & & 1 & 0 & 0 & 0
      \\end{array}

  References
  ----------

  [1] P. E. Kloeden, E. Platen, Numerical Solution of Stochastic Differential
      Equations, 2nd Edition, Springer, Berlin Heidelberg New York, 1995.
  """
  return Wrappers.wrap(Wrappers.KlPl, f=f, g=g, dt=dt, intg_type=intg_type,
                       var_type=var_type, wiener_type=wiener_type, show_code=show_code)



# -*- coding: utf-8 -*-

from brainpy import math
from brainpy.simulation.brainobjects.base import DynamicalSystem
from brainpy.simulation._imports import mjax

__all__ = [
  'Activation'
]


class Activation(DynamicalSystem):
  """Activation Layer.

  Parameters
  ----------
  activation : str
    The name of the activation function.
  name : optional, str
    The name of the class.
  setting : Any
    The settings for the activation function.
  """
  def __init__(self, activation, name=None, **setting):
    if name is None:
      name = self.unique_name(type=f'Activation_{activation}')
    super(Activation, self).__init__(name=name)

    self._activation = mjax.activations.get(activation)
    self._setting = setting

  def update(self, x, **kwargs):
    return self._activation(x, **self._setting)



# -*- coding: utf-8 -*-

import logging

from brainpy import errors
from brainpy.base.base import Base

try:
  import numba
  from brainpy.math.numpy import ast2numba
except ModuleNotFoundError:
  ast2numba = None
  numba = None


__all__ = [
  'jit',
  'vmap',
  'pmap',
]

logger = logging.getLogger('brainpy.math.numpy.compilation')


def jit(obj_or_fun, nopython=True, fastmath=True, parallel=False, nogil=False,
        forceobj=False, looplift=True, error_model='python', inline='never',
        boundscheck=None, show_code=False, debug=False, **kwargs):
  """Just-In-Time (JIT) Compilation in NumPy backend.

  JIT compilation in NumPy backend relies on `Numba <http://numba.pydata.org/>`_. However,
  in BrainPy, `bp.math.numpy.jit()` can apply to class objects, especially the instance
  of :py:class:`brainpy.DynamicalSystem`.

  If you are using JAX backend, please refer to the JIT compilation in
  JAX backend `bp.math.jax.jit() <brainpy.math.jax.jit.rst>`_.

  Parameters
  ----------
  obj_or_fun : callable, Base
    The function or the base model to jit compile.

  nopython : bool
    Set to True to disable the use of PyObjects and Python API
    calls. Default value is True.

  fastmath : bool
    In certain classes of applications strict IEEE 754 compliance
    is less important. As a result it is possible to relax some
    numerical rigour with view of gaining additional performance.
    The way to achieve this behaviour in Numba is through the use
    of the ``fastmath`` keyword argument.

  parallel : bool
    Enables automatic parallelization (and related optimizations) for
    those operations in the function known to have parallel semantics.

  nogil : bool
    Whenever Numba optimizes Python code to native code that only
    works on native types and variables (rather than Python objects),
    it is not necessary anymore to hold Python’s global interpreter
    lock (GIL). Numba will release the GIL when entering such a
    compiled function if you passed ``nogil=True``.

  forceobj: bool
    Set to True to force the use of PyObjects for every value.
    Default value is False.

  looplift: bool
    Set to True to enable jitting loops in nopython mode while
    leaving surrounding code in object mode. This allows functions
    to allocate NumPy arrays and use Python objects, while the
    tight loops in the function can still be compiled in nopython
    mode. Any arrays that the tight loop uses should be created
    before the loop is entered. Default value is True.

  error_model: str
    The error-model affects divide-by-zero behavior.
    Valid values are 'python' and 'numpy'. The 'python' model
    raises exception.  The 'numpy' model sets the result to
    *+/-inf* or *nan*. Default value is 'python'.

  inline: str or callable
    The inline option will determine whether a function is inlined
    at into its caller if called. String options are 'never'
    (default) which will never inline, and 'always', which will
    always inline. If a callable is provided it will be called with
    the call expression node that is requesting inlining, the
    caller's IR and callee's IR as arguments, it is expected to
    return Truthy as to whether to inline.
    NOTE: This inlining is performed at the Numba IR level and is in
    no way related to LLVM inlining.

  boundscheck: bool or None
    Set to True to enable bounds checking for array indices. Out
    of bounds accesses will raise IndexError. The default is to
    not do bounds checking. If False, bounds checking is disabled,
    out of bounds accesses can produce garbage results or segfaults.
    However, enabling bounds checking will slow down typical
    functions, so it is recommended to only use this flag for
    debugging. You can also set the NUMBA_BOUNDSCHECK environment
    variable to 0 or 1 to globally override this flag. The default
    value is None, which under normal execution equates to False,
    but if debug is set to True then bounds checking will be
    enabled.

  show_code : bool
    Debugging.

  kwargs

  Returns
  -------
  res : callable, Base
    The jitted objects.
  """

  # checking
  if ast2numba is None or numba is None:
    raise errors.PackageMissingError('JIT compilation in numpy backend need Numba. '
                                     'Please install numba via: \n'
                                     '>>> pip install numba\n'
                                     '>>> # or \n'
                                     '>>> conda install numba')
  return ast2numba.jit(obj_or_fun, show_code=show_code,
                       nopython=nopython, fastmath=fastmath, parallel=parallel, nogil=nogil,
                       forceobj=forceobj, looplift=looplift, error_model=error_model,
                       inline=inline, boundscheck=boundscheck, debug=debug)


def vmap(obj_or_func, *args, **kwargs):
  """Vectorization Compilation in NumPy backend.

  Vectorization compilation is not implemented in NumPy backend.
  Please refer to the vectorization compilation in JAX backend
  `bp.math.jax.vmap() <brainpy.math.jax.vmap.rst>`_.

  Parameters
  ----------
  obj_or_func
  args
  kwargs

  Returns
  -------

  """
  _msg = 'Vectorize compilation is only supported in JAX backend, not available in numpy backend. \n' \
         'You can switch to JAX backend by `brainpy.math.use_backend("jax")`'
  logger.error(_msg)
  raise errors.BrainPyError(_msg)


def pmap(obj_or_func, *args, **kwargs):
  """Parallel Compilation in NumPy backend.

  Parallel compilation is not implemented in NumPy backend.
  Please refer to the parallel compilation in JAX backend
  `bp.math.jax.pmap() <brainpy.math.jax.pmap.rst>`_.

  Parameters
  ----------
  obj_or_func
  args
  kwargs

  Returns
  -------

  """
  _msg = 'Parallel compilation is only supported in JAX backend, not available in numpy backend. \n' \
         'You can switch to JAX backend by `brainpy.math.use_backend("jax")`'
  logger.error(_msg)
  raise errors.BrainPyError(_msg)



# -*- coding: utf-8 -*-

import numpy as np
import pytest

from brainpy import errors
from brainpy.integrators import ode
from brainpy.integrators import odeint



def test_exp_euler():
    method = 'exponential_euler'

    print(f'Test {method} method:')
    print()

    def func(m, t, V):
        alpha = 0.1 * (V + 40) / (1 - np.exp(-(V + 40) / 10))
        beta = 4.0 * np.exp(-(V + 65) / 18)
        dmdt = alpha * (1 - m) - beta * m
        return dmdt

    odeint(method=method, show_code=True, f=func)

    with pytest.raises(errors.CodeError):
        def func(f, t, V):
            alpha = 0.1 * (V + 40) / (1 - np.exp(-(V + 40) / 10))
            beta = 4.0 * np.exp(-(V + 65) / 18)
            dmdt = alpha * (1 - f) - beta * f
            return dmdt
        odeint(method=method, show_code=True, f=func)

    with pytest.raises(errors.CodeError):
        def func(m, t, dt):
            alpha = 0.1 * (dt + 40) / (1 - np.exp(-(dt + 40) / 10))
            beta = 4.0 * np.exp(-(dt + 65) / 18)
            dmdt = alpha * (1 - m) - beta * m
            return dmdt
        odeint(method=method, show_code=True, f=func)

    with pytest.raises(errors.CodeError):
        def func(m, t, m_new):
            alpha = 0.1 * (m_new + 40) / (1 - np.exp(-(m_new + 40) / 10))
            beta = 4.0 * np.exp(-(m_new + 65) / 18)
            dmdt = alpha * (1 - m) - beta * m
            return dmdt
        odeint(method=method, show_code=True, f=func)

    with pytest.raises(errors.CodeError):
        def func(m, t, exp):
            alpha = 0.1 * (exp + 40) / (1 - np.exp(-(exp + 40) / 10))
            beta = 4.0 * np.exp(-(exp + 65) / 18)
            dmdt = alpha * (1 - m) - beta * m
            return dmdt
        odeint(method=method, show_code=True, f=func)



    print('-' * 40)


if __name__ == '__main__':
    test_exp_euler()



# -*- coding: utf-8 -*-


from .activation import *
from .attention import *
from .conv import *
from .dense import *
from .dropout import *
from .normalization import *
from .pooling import *
from .readout import *
from .recurrent import *
from .sequential import *



# -*- coding: utf-8 -*-

"""This module aims to provide commonly used methods for simulated neuronal data.

"""


import numpy as np

from brainpy import tools, math

__all__ = [
  'cross_correlation',
  'voltage_fluctuation',
  'raster_plot',
  'firing_rate',
]


@tools.numba_jit
def _cc(states, i, j):
  sqrt_ij = np.sqrt(np.sum(states[i]) * np.sum(states[j]))
  k = 0. if sqrt_ij == 0. else np.sum(states[i] * states[j]) / sqrt_ij
  return k


def cross_correlation(spikes, bin, dt=None):
  r"""Calculate cross correlation index between neurons.

  The coherence [1]_ between two neurons i and j is measured by their
  cross-correlation of spike trains at zero time lag within a time bin
  of :math:`\Delta t = \tau`. More specifically, suppose that a long
  time interval T is divided into small bins of :math:`\Delta t` and
  that two spike trains are given by :math:`X(l)=` 0 or 1, :math:`Y(l)=` 0
  or 1, :math:`l=1,2, \ldots, K(T / K=\tau)`. Thus, we define a coherence
  measure for the pair as:

  .. math::

      \kappa_{i j}(\tau)=\frac{\sum_{l=1}^{K} X(l) Y(l)}
      {\sqrt{\sum_{l=1}^{K} X(l) \sum_{l=1}^{K} Y(l)}}

  The population coherence measure :math:`\kappa(\tau)` is defined by the
  average of :math:`\kappa_{i j}(\tau)` over many pairs of neurons in the
  network.

  Parameters
  ----------
  spikes :
      The history of spike states of the neuron group.
      It can be easily get via `StateMonitor(neu, ['spike'])`.
  bin : float, int
      The time bin to normalize spike states.
  dt : float, optional
      The time precision.

  Returns
  -------
  cc_index : float
      The cross correlation value which represents the synchronization index.

  References
  ----------
  .. [1] Wang, Xiao-Jing, and György Buzsáki. "Gamma oscillation by synaptic
         inhibition in a hippocampal interneuronal network model." Journal of
         neuroscience 16.20 (1996): 6402-6413.
  """

  dt = math.get_dt() if dt is None else dt
  bin_size = int(bin / dt)
  num_hist, num_neu = spikes.shape
  num_bin = int(np.ceil(num_hist / bin_size))
  if num_bin * bin_size != num_hist:
    spikes = np.append(spikes, np.zeros((num_bin * bin_size - num_hist, num_neu)), axis=0)
  states = spikes.T.reshape((num_neu, num_bin, bin_size))
  states = (np.sum(states, axis=2) > 0.).astype(np.float_)
  all_k = []
  for i in range(num_neu):
    for j in range(i + 1, num_neu):
      all_k.append(_cc(states, i, j))
  return np.mean(all_k)


@tools.numba_jit
def _var(neu_signal):
  return np.mean(neu_signal * neu_signal) - np.mean(neu_signal) ** 2


def voltage_fluctuation(potentials):
  r"""Calculate neuronal synchronization via voltage variance.

  The method comes from [1]_ [2]_ [3]_.

  First, average over the membrane potential :math:`V`

  .. math::

      V(t) = \frac{1}{N} \sum_{i=1}^{N} V_i(t)

  The variance of the time fluctuations of :math:`V(t)` is

  .. math::

      \sigma_V^2 = \left\langle \left[ V(t) \right]^2 \right\rangle_t -
      \left[ \left\langle V(t) \right\rangle_t \right]^2

  where :math:`\left\langle \ldots \right\rangle_t = (1 / T_m) \int_0^{T_m} dt \, \ldots`
  denotes time-averaging over a large time, :math:`\tau_m`. After normalization
  of :math:`\sigma_V` to the average over the population of the single cell
  membrane potentials

  .. math::

      \sigma_{V_i}^2 = \left\langle\left[ V_i(t) \right]^2 \right\rangle_t -
      \left[ \left\langle V_i(t) \right\rangle_t \right]^2

  one defines a synchrony measure, :math:`\chi (N)`, for the activity of a system
  of :math:`N` neurons by:

  .. math::

      \chi^2 \left( N \right) = \frac{\sigma_V^2}{ \frac{1}{N} \sum_{i=1}^N
      \sigma_{V_i}^2}

  Parameters
  ----------
  potentials :
      The membrane potential matrix of the neuron group.

  Returns
  -------
  sync_index : float
      The synchronization index.

  References
  ----------
  .. [1] Golomb, D. and Rinzel J. (1993) Dynamics of globally coupled
         inhibitory neurons with heterogeneity. Phys. Rev. reversal_potential 48:4810-4814.
  .. [2] Golomb D. and Rinzel J. (1994) Clustering in globally coupled
         inhibitory neurons. Physica D 72:259-282.
  .. [3] David Golomb (2007) Neuronal synchrony measures. Scholarpedia, 2(1):1347.
  """

  num_hist, num_neu = potentials.shape
  avg = np.mean(potentials, axis=1)
  avg_var = np.mean(avg * avg) - np.mean(avg) ** 2
  neu_vars = []
  for i in range(num_neu):
    neu_vars.append(_var(potentials[:, i]))
  var_mean = np.mean(neu_vars)
  return avg_var / var_mean if var_mean != 0. else 1.


def raster_plot(sp_matrix, times):
  """Get spike raster plot which displays the spiking activity
  of a group of neurons over time.

  Parameters
  ----------
  sp_matrix : bnp.ndarray
      The matrix which record spiking activities.
  times : bnp.ndarray
      The time steps.

  Returns
  -------
  raster_plot : tuple
      Include (neuron index, spike time).
  """
  elements = np.where(sp_matrix > 0.)
  index = elements[1]
  time = times[elements[0]]
  return index, time


def firing_rate(sp_matrix, width, dt=None):
  r"""Calculate the mean firing rate over in a neuron group.

  This method is adopted from Brian2.

  The firing rate in trial :math:`k` is the spike count :math:`n_{k}^{sp}`
  in an interval of duration :math:`T` divided by :math:`T`:

  .. math::

      v_k = {n_k^{sp} \over T}

  Parameters
  ----------
  sp_matrix : bnp.ndarray
    The spike matrix which record spiking activities.
  width : int, float
    The width of the ``window`` in millisecond.
  dt : float, optional
    The sample rate.

  Returns
  -------
  rate : numpy.ndarray
      The population rate in Hz, smoothed with the given window.
  """
  rate = math.sum(sp_matrix, axis=1) / sp_matrix.shape[1]
  dt = math.get_dt() if dt is None else dt
  width1 = int(width / 2 / dt) * 2 + 1
  window = math.ones(width1) * 1000 / width
  return np.convolve(rate, window, mode='same')



# -*- coding: utf-8 -*-

import numpy

from brainpy.tools import numba_jit

__all__ = [
  # math funcs
  'real', 'imag', 'conj', 'conjugate', 'ndim', 'isreal', 'isscalar',
  'add', 'reciprocal', 'negative', 'positive', 'multiply', 'divide',
  'power', 'subtract', 'true_divide', 'floor_divide', 'float_power',
  'fmod', 'mod', 'modf', 'divmod', 'remainder', 'abs', 'exp', 'exp2',
  'expm1', 'log', 'log10', 'log1p', 'log2', 'logaddexp', 'logaddexp2',
  'lcm', 'gcd', 'arccos', 'arccosh', 'arcsin', 'arcsinh', 'arctan',
  'arctan2', 'arctanh', 'cos', 'cosh', 'sin', 'sinc', 'sinh', 'tan',
  'tanh', 'deg2rad', 'hypot', 'rad2deg', 'degrees', 'radians', 'round', 'around',
  'round_', 'rint', 'floor', 'ceil', 'trunc', 'fix', 'prod', 'sum', 'diff',
  'median', 'nancumprod', 'nancumsum', 'nanprod', 'nansum', 'cumprod',
  'cumsum', 'ediff1d', 'cross', 'trapz', 'isfinite', 'isinf', 'isnan',
  'signbit', 'copysign', 'nextafter', 'ldexp', 'frexp', 'convolve',
  'sqrt', 'cbrt', 'square', 'absolute', 'fabs', 'sign', 'heaviside',
  'maximum', 'minimum', 'fmax', 'fmin', 'interp', 'clip', 'angle',

  # Elementwise bit operations
  'bitwise_and', 'bitwise_not', 'bitwise_or', 'bitwise_xor',
  'invert', 'left_shift', 'right_shift',

  # logic funcs
  'equal', 'not_equal', 'greater', 'greater_equal', 'less', 'less_equal',
  'array_equal', 'isclose', 'allclose', 'logical_and', 'logical_not',
  'logical_or', 'logical_xor', 'all', 'any',

  # array manipulation
  'shape', 'size', 'reshape', 'ravel', 'moveaxis', 'transpose', 'swapaxes',
  'concatenate', 'stack', 'vstack', 'hstack', 'dstack', 'column_stack',
  'split', 'dsplit', 'hsplit', 'vsplit', 'tile', 'repeat', 'unique',
  'append', 'flip', 'fliplr', 'flipud', 'roll', 'atleast_1d', 'atleast_2d',
  'atleast_3d', 'expand_dims', 'squeeze', 'sort', 'argsort', 'argmax', 'argmin',
  'argwhere', 'nonzero', 'flatnonzero', 'where', 'searchsorted', 'extract',
  'count_nonzero', 'max', 'min',

  # array creation
  'empty', 'empty_like', 'ones', 'ones_like', 'zeros', 'zeros_like', 'full',
  'full_like', 'eye', 'identity', 'array', 'asarray', 'arange', 'linspace',
  'logspace', 'meshgrid', 'diag', 'tri', 'tril', 'triu', 'vander', 'fill_diagonal',

  # indexing funcs
  'nonzero', 'where', 'tril_indices', 'tril_indices_from', 'triu_indices',
  'triu_indices_from', 'take', 'diag', 'select',

  # statistic funcs
  'nanmin', 'nanmax', 'ptp', 'percentile', 'nanpercentile', 'quantile', 'nanquantile',
  'median', 'average', 'mean', 'std', 'var', 'nanmedian', 'nanmean', 'nanstd', 'nanvar',
  'corrcoef', 'correlate', 'cov', 'histogram', 'bincount', 'digitize',

  # window funcs
  'bartlett', 'blackman', 'hamming', 'hanning', 'kaiser',

  # constants
  'e', 'pi', 'inf',

  # linear algebra
  'dot', 'vdot', 'inner', 'outer', 'kron', 'matmul', 'trace',

  # data types
  'dtype', 'finfo', 'iinfo', 'bool_', 'uint8', 'uint16', 'uint32', 'uint64',
  'int_', 'int8', 'int16', 'int32', 'int64', 'float_', 'float16', 'float32', 'float64',
  'complex_', 'complex64', 'complex128', 'set_int_', 'set_float_', 'set_complex_',

  # others
  'take_along_axis', 'clip_by_norm',
]

# math funcs
# ----------

isreal = numpy.isreal
isscalar = numpy.isscalar
real = numpy.real
imag = numpy.imag
conj = numpy.conj
conjugate = numpy.conjugate
ndim = numpy.ndim

add = numpy.add
reciprocal = numpy.reciprocal
negative = numpy.negative
positive = numpy.positive
multiply = numpy.multiply
divide = numpy.divide
power = numpy.power
subtract = numpy.subtract
true_divide = numpy.true_divide
floor_divide = numpy.floor_divide
float_power = numpy.float_power
fmod = numpy.fmod
mod = numpy.mod
modf = numpy.modf
divmod = numpy.divmod
remainder = numpy.remainder
abs = numpy.absolute

exp = numpy.exp
exp2 = numpy.exp2
expm1 = numpy.expm1
log = numpy.log
log10 = numpy.log10
log1p = numpy.log1p
log2 = numpy.log2
logaddexp = numpy.logaddexp
logaddexp2 = numpy.logaddexp2

lcm = numpy.lcm
gcd = numpy.gcd

arccos = numpy.arccos
arccosh = numpy.arccosh
arcsin = numpy.arcsin
arcsinh = numpy.arcsinh
arctan = numpy.arctan
arctan2 = numpy.arctan2
arctanh = numpy.arctanh
cos = numpy.cos
cosh = numpy.cosh
sin = numpy.sin
sinc = numpy.sinc
sinh = numpy.sinh
tan = numpy.tan
tanh = numpy.tanh
deg2rad = numpy.deg2rad
hypot = numpy.hypot
rad2deg = numpy.rad2deg
degrees = numpy.degrees
radians = numpy.radians

round = numpy.round
around = numpy.around
round_ = numpy.round_
rint = numpy.rint
floor = numpy.floor
ceil = numpy.ceil
trunc = numpy.trunc
fix = numpy.fix

prod = numpy.prod
sum = numpy.sum
diff = numpy.diff
median = numpy.median
nancumprod = numpy.nancumprod
nancumsum = numpy.nancumsum
nanprod = numpy.nanprod
nansum = numpy.nansum
cumprod = numpy.cumprod
cumsum = numpy.cumsum
ediff1d = numpy.ediff1d
cross = numpy.cross
trapz = numpy.trapz

isfinite = numpy.isfinite
isinf = numpy.isinf
isnan = numpy.isnan
signbit = numpy.signbit
copysign = numpy.copysign
nextafter = numpy.nextafter
ldexp = numpy.ldexp
frexp = numpy.frexp

convolve = numpy.convolve
sqrt = numpy.sqrt
cbrt = numpy.cbrt
square = numpy.square
absolute = numpy.absolute
fabs = numpy.fabs
sign = numpy.sign
heaviside = numpy.heaviside
maximum = numpy.maximum
minimum = numpy.minimum
fmax = numpy.fmax
fmin = numpy.fmin
interp = numpy.interp
clip = numpy.clip
angle = numpy.angle

# binary funcs
# -------------

bitwise_and = numpy.bitwise_and
bitwise_not = numpy.bitwise_not
bitwise_or = numpy.bitwise_or
bitwise_xor = numpy.bitwise_xor
invert = numpy.invert
left_shift = numpy.left_shift
right_shift = numpy.right_shift

# logic funcs
# -----------

equal = numpy.equal
not_equal = numpy.not_equal
greater = numpy.greater
greater_equal = numpy.greater_equal
less = numpy.less
less_equal = numpy.less_equal
array_equal = numpy.array_equal
isclose = numpy.isclose
allclose = numpy.allclose

logical_and = numpy.logical_and
logical_not = numpy.logical_not
logical_or = numpy.logical_or
logical_xor = numpy.logical_xor

all = numpy.all
any = numpy.any

# array manipulation
# ------------------

shape = numpy.shape
size = numpy.size
reshape = numpy.reshape
ravel = numpy.ravel
moveaxis = numpy.moveaxis
transpose = numpy.transpose
swapaxes = numpy.swapaxes
concatenate = numpy.concatenate
stack = numpy.stack
vstack = numpy.vstack
hstack = numpy.hstack
dstack = numpy.dstack
column_stack = numpy.column_stack
split = numpy.split
dsplit = numpy.dsplit
hsplit = numpy.hsplit
vsplit = numpy.vsplit
tile = numpy.tile
repeat = numpy.repeat
unique = numpy.unique
append = numpy.append
flip = numpy.flip
fliplr = numpy.fliplr
flipud = numpy.flipud
roll = numpy.roll
atleast_1d = numpy.atleast_1d
atleast_2d = numpy.atleast_2d
atleast_3d = numpy.atleast_3d
expand_dims = numpy.expand_dims
squeeze = numpy.squeeze
sort = numpy.sort
argsort = numpy.argsort
argmax = numpy.argmax
argmin = numpy.argmin
argwhere = numpy.argwhere
nonzero = numpy.nonzero
flatnonzero = numpy.flatnonzero
where = numpy.where
searchsorted = numpy.searchsorted
extract = numpy.extract
count_nonzero = numpy.count_nonzero
max = numpy.max
min = numpy.min

# array creation
# --------------

ndarray = numpy.ndarray
empty = numpy.empty
empty_like = numpy.empty_like
ones = numpy.ones
ones_like = numpy.ones_like
zeros = numpy.zeros
zeros_like = numpy.zeros_like
full = numpy.full
full_like = numpy.full_like
eye = numpy.eye
identity = numpy.identity

array = numpy.array
asarray = numpy.asarray

arange = numpy.arange
linspace = numpy.linspace
logspace = numpy.logspace
meshgrid = numpy.meshgrid

diag = numpy.diag
tri = numpy.tri
tril = numpy.tril
triu = numpy.triu
vander = numpy.vander


def fill_diagonal(a, val, wrap=False):
  numpy.fill_diagonal(a, val, wrap=wrap)
  return a


# indexing funcs
# --------------

tril_indices = numpy.tril_indices
tril_indices_from = numpy.tril_indices_from
triu_indices = numpy.triu_indices
triu_indices_from = numpy.triu_indices_from
take = numpy.take
select = numpy.select

# statistic funcs
# ---------------
nanmin = numpy.nanmin
nanmax = numpy.nanmax
ptp = numpy.ptp
percentile = numpy.percentile
nanpercentile = numpy.nanpercentile
quantile = numpy.quantile
nanquantile = numpy.nanquantile

average = numpy.average
mean = numpy.mean
std = numpy.std
var = numpy.var
nanmedian = numpy.nanmedian
nanmean = numpy.nanmean
nanstd = numpy.nanstd
nanvar = numpy.nanvar

corrcoef = numpy.corrcoef
correlate = numpy.correlate
cov = numpy.cov

histogram = numpy.histogram
bincount = numpy.bincount
digitize = numpy.digitize

bartlett = numpy.bartlett
blackman = numpy.blackman
hamming = numpy.hamming
hanning = numpy.hanning
kaiser = numpy.kaiser

# constants
# ---------

e = numpy.e
pi = numpy.pi
inf = numpy.inf

# linear algebra
# --------------

dot = numpy.dot
vdot = numpy.vdot
inner = numpy.inner
outer = numpy.outer
kron = numpy.kron
matmul = numpy.matmul
trace = numpy.trace

# data types
# ----------

dtype = numpy.dtype
finfo = numpy.finfo
iinfo = numpy.iinfo

bool_ = numpy.bool_
uint8 = numpy.uint8
uint16 = numpy.uint16
uint32 = numpy.uint32
uint64 = numpy.uint64
int_ = numpy.int_
int8 = numpy.int8
int16 = numpy.int16
int32 = numpy.int32
int64 = numpy.int64
float_ = numpy.float_
float16 = numpy.float16
float32 = numpy.float32
float64 = numpy.float64
complex_ = numpy.complex_
complex64 = numpy.complex64
complex128 = numpy.complex128


def set_int_(int_type):
  """Set the default ``int`` type.

  Parameters
  ----------
  int_type : type
  """
  global int_
  assert isinstance(int_type, type)
  int_ = int_type


def set_float_(float_type):
  """Set the default ``float`` type.

  Parameters
  ----------
  float_type : type
  """
  global float_
  assert isinstance(float_type, type)
  float_ = float_type


def set_complex_(complex_type):
  """Set the default ``complex`` type.

  Parameters
  ----------
  complex_type : type
  """
  global complex_
  assert isinstance(complex_type, type)
  complex_ = complex_type


# others
# -------

take_along_axis = numpy.take_along_axis


@numba_jit
def clip_by_norm(t, clip_norm, axis=None):
  """Clips values to a maximum L2-norm.

  Given a tensor ``t``, and a maximum clip value ``clip_norm``, this operation
  normalizes ``t`` so that its L2-norm is less than or equal to ``clip_norm``,
  along the dimensions given in ``axis``. Specifically, in the default case
  where all dimensions are used for calculation, if the L2-norm of ``t`` is
  already less than or equal to ``clip_norm``, then ``t`` is not modified. If
  the L2-norm is greater than ``clip_norm``, then this operation returns a
  tensor of the same type and shape as ``t`` with its values set to:

  .. math::
    t * clip_norm / l2norm(t)

  In this case, the L2-norm of the output tensor is `clip_norm`.

  As another example, if ``t`` is a matrix and ``axis=1``, then each row
  of the output will have L2-norm less than or equal to ``clip_norm``. If
  ``axis=0`` instead, each column of the output will be clipped.

  This operation is typically used to clip gradients before applying them with
  an optimizer.
  """
  l2norm = sqrt(sum(t * t, axis=axis, keepdims=True))
  clip_values = t * clip_norm / maximum(l2norm, clip_norm)
  return clip_values



# -*- coding: utf-8 -*-

"""
This module provides numerical solvers for various differential equations.


"""


from . import dde
from . import fde
from . import ode
from . import sde
from .constants import *
from .delay_vars import *
from .wrapper import *



# -*- coding: utf-8 -*-

import abc

from brainpy import math

__all__ = [
  'Initializer',
]


class Initializer(abc.ABC):
  """Base Initialization Class."""
  def __init__(self, dtype=None):
    self.dtype = math.float_ if dtype is None else dtype

  @abc.abstractmethod
  def __call__(self, shape):
    raise NotImplementedError



# -*- coding: utf-8 -*-

from . import ode
from . import sde

__all__ = [
  'SUPPORTED_ODE_METHODS',
  'SUPPORTED_SDE_METHODS',

  'odeint',
  'sdeint',
  'ddeint',
  'fdeint',

  'set_default_odeint', 'get_default_odeint',
  'set_default_sdeint', 'get_default_sdeint',
]

_DEFAULT_ODE_METHOD = 'euler'
_DEFAULT_SDE_METHOD = 'euler'
SUPPORTED_ODE_METHODS = [m for m in dir(ode) if not m.startswith('__') and callable(getattr(ode, m))]
SUPPORTED_SDE_METHODS = [m for m in dir(sde) if not m.startswith('__') and callable(getattr(sde, m))]


def _wrapper(f, module, method, **kwargs):
  integrator = getattr(module, method)
  return integrator(f, **kwargs)


def odeint(f=None, method=None, **kwargs):
  """Numerical integration for ODE_INT.

  Parameters
  ----------
  f : callable
  method : str
  kwargs :

  Returns
  -------
  int_f : callable
      The numerical solver of `f`.
  """
  if method is None:
    method = _DEFAULT_ODE_METHOD
  if method not in SUPPORTED_ODE_METHODS:
    raise ValueError(f'Unknown ODE_INT numerical method "{method}". Currently '
                     f'BrainPy only support: {SUPPORTED_ODE_METHODS}')

  if f is None:
    return lambda f: _wrapper(f, method=method, module=ode, **kwargs)
  else:
    return _wrapper(f, method=method, module=ode, **kwargs)


def sdeint(f=None, method=None, **kwargs):
  if method is None:
    method = _DEFAULT_SDE_METHOD
  if method not in SUPPORTED_SDE_METHODS:
    raise ValueError(f'Unknown SDE_INT numerical method "{method}". Currently '
                     f'BrainPy only support: {SUPPORTED_SDE_METHODS}')

  if f is None:
    return lambda f: _wrapper(f, method=method, module=sde, **kwargs)
  else:
    return _wrapper(f, method=method, module=sde, **kwargs)


def ddeint():
  raise NotImplementedError


def fdeint():
  raise NotImplementedError


def set_default_odeint(method):
  """Set the default ODE_INT numerical integrator method for differential equations.

  Parameters
  ----------
  method : str, callable
      Numerical integrator method.
  """
  if not isinstance(method, str):
    raise ValueError(f'Only support string, not {type(method)}.')
  if method not in SUPPORTED_ODE_METHODS:
    raise ValueError(f'Unsupported ODE_INT numerical method: {method}.')

  global _DEFAULT_ODE_METHOD
  _DEFAULT_ODE_METHOD = method


def get_default_odeint():
  """Get the default ODE_INT numerical integrator method.

  Returns
  -------
  method : str
      The default numerical integrator method.
  """
  return _DEFAULT_ODE_METHOD


def set_default_sdeint(method):
  """Set the default SDE_INT numerical integrator method for differential equations.

  Parameters
  ----------
  method : str, callable
      Numerical integrator method.
  """
  if not isinstance(method, str):
    raise ValueError(f'Only support string, not {type(method)}.')
  if method not in SUPPORTED_SDE_METHODS:
    raise ValueError(f'Unsupported SDE_INT numerical method: {method}.')

  global _DEFAULT_SDE_METHOD
  _DEFAULT_SDE_METHOD = method


def get_default_sdeint():
  """Get the default ODE_INT numerical integrator method.

  Returns
  -------
  method : str
      The default numerical integrator method.
  """
  return _DEFAULT_SDE_METHOD



# -*- coding: utf-8 -*-

import matplotlib.pyplot as plt
import pytest

import brainpy as bp


def test_random_prob():
  for bk in ['numpy', 'jax']:
    bp.math.use_backend(bk)

    conn1 = bp.connect.FixedProb(prob=0.1, seed=123)
    conn1(pre_size=(10, 20), post_size=(10, 20))
    pre_ids, post_ids = conn1.require('pre_ids', 'post_ids')

    conn2 = bp.connect.FixedProb(prob=0.1, seed=123)
    conn2(pre_size=(10, 20), post_size=(10, 20))
    mat = conn2.require(bp.connect.CONN_MAT)
    pre_ids2, post_ids2 = bp.math.where(mat)


def test_random_fix_pre1():
  for bk in ['numpy', 'jax']:
    bp.math.use_backend(bk)

    for num in [0.4, 20]:
      conn1 = bp.connect.FixedPreNum(num, seed=1234)(pre_size=(10, 15), post_size=(10, 20))
      mat1 = conn1.require(bp.connect.CONN_MAT)

      conn2 = bp.connect.FixedPreNum(num, seed=1234)(pre_size=(10, 15), post_size=(10, 20))
      mat2 = conn2.require(bp.connect.CONN_MAT)

      assert bp.math.array_equal(mat1, mat2)


def test_random_fix_pre2():
  for bk in ['numpy', 'jax']:
    bp.math.use_backend(bk)

    for num in [0.5, 3]:
      conn1 = bp.connect.FixedPreNum(num, seed=1234)(pre_size=5, post_size=4)
      mat1 = conn1.require(bp.connect.CONN_MAT)
      print(mat1)


def test_random_fix_pre3():
  bp.math.use_backend('numpy')
  conn1 = bp.connect.FixedPreNum(num=6, seed=1234)(pre_size=3, post_size=4)
  with pytest.raises(AssertionError):
    conn1.require(bp.connect.CONN_MAT)


def test_random_fix_post1():
  for bk in ['numpy', 'jax']:
    bp.math.use_backend(bk)

    for num in [0.4, 20]:
      conn1 = bp.connect.FixedPostNum(num, seed=1234)(pre_size=(10, 15), post_size=(10, 20))
      mat1 = conn1.require(bp.connect.CONN_MAT)

      conn2 = bp.connect.FixedPostNum(num, seed=1234)(pre_size=(10, 15), post_size=(10, 20))
      mat2 = conn2.require(bp.connect.CONN_MAT)

      assert bp.math.array_equal(mat1, mat2)


def test_random_fix_post2():
  for bk in ['numpy', 'jax']:
    bp.math.use_backend(bk)

    for num in [0.5, 3]:
      conn1 = bp.connect.FixedPostNum(num, seed=1234)(pre_size=5, post_size=4)
      mat1 = conn1.require(bp.connect.CONN_MAT)
      print(mat1)


def test_random_fix_post3():
  bp.math.use_backend('numpy')
  conn1 = bp.connect.FixedPostNum(num=6, seed=1234)(pre_size=3, post_size=4)
  with pytest.raises(AssertionError):
    conn1.require(bp.connect.CONN_MAT)


def test_gaussian_prob1():
  bp.math.use_backend('numpy')
  conn = bp.connect.GaussianProb(sigma=1.)(pre_size=100)
  mat = conn.require(bp.connect.CONN_MAT)

  # plt.imshow(mat)
  # plt.show()


def test_gaussian_prob2():
  bp.math.use_backend('numpy')

  conn = bp.connect.GaussianProb(sigma=4)(pre_size=(50, 50))
  mat = conn.require(bp.connect.CONN_MAT)

  # plt.imshow(mat)
  # plt.show()

  # plt.imshow(mat[0].reshape(conn.pre_size))
  # plt.show()


def test_gaussian_prob3():
  bp.math.use_backend('numpy')
  conn = bp.connect.GaussianProb(sigma=4, periodic_boundary=True)(pre_size=(50, 50))
  mat = conn.require(bp.connect.CONN_MAT)

  # plt.imshow(mat)
  # plt.show()

  # plt.imshow(mat[0].reshape(conn.pre_size))
  # plt.show()


def test_gaussian_prob4():
  bp.math.use_backend('numpy')
  conn = bp.connect.GaussianProb(sigma=4, periodic_boundary=True)(pre_size=(20, 20, 20))
  conn.require(bp.connect.CONN_MAT,
               bp.connect.PRE_IDS, bp.connect.POST_IDS,
               bp.connect.PRE2POST, bp.connect.POST_IDS)


def test_SmallWorld1():
  for bk in ['numpy', 'jax']:
    bp.math.use_backend(bk)

    conn = bp.connect.SmallWorld(num_neighbor=2, prob=0.5)
    conn(pre_size=(20, 3), post_size=(20, 3))
    with pytest.raises(NotImplementedError):
      conn.require(bp.connect.CONN_MAT)


def test_SmallWorld2():
  for bk in ['numpy', 'jax']:
    bp.math.use_backend(bk)

    conn = bp.connect.SmallWorld(num_neighbor=2, prob=0.5)
    conn(pre_size=(100,), post_size=(100,))
    conn.require(bp.connect.CONN_MAT,
                 bp.connect.PRE_IDS, bp.connect.POST_IDS,
                 bp.connect.PRE2POST, bp.connect.POST_IDS)


def test_ScaleFreeBA():
  for bk in ['numpy', 'jax']:
    bp.math.use_backend(bk)

    conn = bp.connect.ScaleFreeBA(m=2)
    for size in [100, (10, 20), (2, 10, 20)]:
      conn(pre_size=size, post_size=size)
      conn.require(bp.connect.CONN_MAT,
                   bp.connect.PRE_IDS, bp.connect.POST_IDS,
                   bp.connect.PRE2POST, bp.connect.POST_IDS)


def test_ScaleFreeBADual():
  for bk in ['numpy', 'jax']:
    bp.math.use_backend(bk)

    conn = bp.connect.ScaleFreeBADual(m1=2, m2=3, p=0.4)
    for size in [100, (10, 20), (2, 10, 20)]:
      conn(pre_size=size, post_size=size)
      conn.require(bp.connect.CONN_MAT,
                   bp.connect.PRE_IDS, bp.connect.POST_IDS,
                   bp.connect.PRE2POST, bp.connect.POST_IDS)


def test_PowerLaw():
  for bk in ['numpy', 'jax']:
    bp.math.use_backend(bk)

    conn = bp.connect.PowerLaw(m=3, p=0.4)
    for size in [100, (10, 20), (2, 10, 20)]:
      conn(pre_size=size, post_size=size)
      conn.require(bp.connect.CONN_MAT,
                   bp.connect.PRE_IDS, bp.connect.POST_IDS,
                   bp.connect.PRE2POST, bp.connect.POST_IDS)



# -*- coding: utf-8 -*-


__all__ = [
  # area.py
  'BrainArea',
  # base.py
  'DynamicalSystem', 'Container',
  # delays.py
  'Delay', 'ConstantDelay',
  # input.py
  'SpikeTimeInput', 'PoissonInput', 'ConstantInput',
  # molecular.py
  'Molecular',
  # network.py
  'Network',
  # neuron.py
  'NeuGroup', 'Channel', 'CondNeuGroup', 'Soma', 'Dendrite',
  # synapse.py
  'TwoEndConn',
]

from .area import *
from .base import *
from .delays import *
from .input import *
from .molecular import *
from .network import *
from .neuron import *
from .synapse import *



# -*- coding: utf-8 -*-

import logging

import matplotlib.pyplot as plt
import numpy as np

from brainpy import errors, math
from brainpy.analysis import stability
from brainpy.analysis.symbolic import base, utils
from brainpy.analysis.symbolic.trajectory import Trajectory

logger = logging.getLogger('brainpy.analysis.symbolic')

__all__ = [
  'PhasePlane',
  '_PhasePlane1D',
  '_PhasePlane2D',
]


class PhasePlane(object):
  """A tool class for phase plane analyzer.

  `PhasePlaneAnalyzer` is used to analyze the phase portrait of 1D
  or 2D dynamical systems. It can also be used to analyze the phase
  portrait of high-dimensional system but with the fixation of other
  variables to preserve only one/two dynamical variables.

  Parameters
  ----------
  integrals : NeuType
      The neuron model which defines the differential equations by using
      `brainpy.integrate`.
  target_vars : dict
      The target variables to analyze, with the format of
      `{'var1': [var_min, var_max], 'var2': [var_min, var_max]}`.
  fixed_vars : dict, optional
      The fixed variables, which means the variables will not be updated.
  pars_update : dict, optional
      The parameters in the differential equations to update.
  numerical_resolution : float, dict, optional
      The variable resolution for numerical iterative solvers.
      This variable will be useful in the solving of nullcline and fixed points
      by using the iterative optimization method. It can be a float, which will
      be used as ``numpy.arange(var_min, var_max, resolution)``. Or, it can be
      a dict, with the format of ``{'var1': resolution1, 'var2': resolution2}``.
      Or, it can be a dict with the format of ``{'var1': np.arange(x, x, x),
      'var2': np.arange(x, x, x)}``.

  options : dict, optional
      The other setting parameters, which includes:

          lim_scale
              float. The axis limit scale factor. Default is 1.05. The setting means
              the axes will be clipped to ``[var_min * (1-lim_scale)/2, var_max * (var_max-1)/2]``.
          sympy_solver_timeout
              float, with the unit of second. The maximum  time allowed to use sympy solver
              to get the variable relationship.
          escape_sympy_solver
              bool. Whether escape to use sympy solver, and directly use numerical optimization
              method to solve the nullcline and fixed points.
          shgo_args
              dict. Arguments of `shgo` optimization method, which can be used to set the
              fields of: constraints, n, iters, callback, minimizer_kwargs, options,
              sampling_method.
          show_shgo
              bool. whether print the shgo's value.
          perturbation
              float. The small perturbation used to solve the function derivative.
          fl_tol
              float. The tolerance of the function value to recognize it as a condidate of
              function root point.
          xl_tol
              float. The tolerance of the l2 norm distances between this point and previous
              points. If the norm distances are all bigger than `xl_tol` means this
              point belong to a new function root point.
  """

  def __init__(
      self,
      integrals,
      target_vars,
      fixed_vars=None,
      pars_update=None,
      numerical_resolution=0.1,
      options=None,
  ):
    # check "model"
    self.model = utils.transform_integrals_to_model(integrals)

    # check "target_vars"
    if not isinstance(target_vars, dict):
      raise errors.BrainPyError('"target_vars" must a dict with the format of: '
                                '{"Variable A": [A_min, A_max], "Variable B": [B_min, B_max]}')
    self.target_vars = target_vars

    # check "fixed_vars"
    if fixed_vars is None:
      fixed_vars = dict()
    if not isinstance(fixed_vars, dict):
      raise errors.BrainPyError('"fixed_vars" must be a dict with the format of: '
                                '{"Variable A": A_value, "Variable B": B_value}')
    self.fixed_vars = fixed_vars

    # check "pars_update"
    if pars_update is None:
      pars_update = dict()
    if not isinstance(pars_update, dict):
      raise errors.BrainPyError('"pars_update" must be a dict with the format of: '
                                '{"Par A": A_value, "Par B": B_value}')
    for key in pars_update.keys():
      if (key not in self.model.scopes) and (key not in self.model.parameters):
        raise errors.BrainPyError(f'"{key}" is not a valid parameter in "{integrals}" model.')
    self.pars_update = pars_update

    # analyzer
    if len(target_vars) == 1:
      self.analyzer = _PhasePlane1D(model_or_integrals=self.model,
                                    target_vars=target_vars,
                                    fixed_vars=fixed_vars,
                                    pars_update=pars_update,
                                    numerical_resolution=numerical_resolution,
                                    options=options)
    elif len(target_vars) == 2:
      self.analyzer = _PhasePlane2D(model_or_integrals=self.model,
                                    target_vars=target_vars,
                                    fixed_vars=fixed_vars,
                                    pars_update=pars_update,
                                    numerical_resolution=numerical_resolution,
                                    options=options)
    else:
      raise errors.BrainPyError('BrainPy only support 1D/2D phase plane analysis. '
                                'Or, you can set "fixed_vars" to fix other variables, '
                                'then make 1D/2D phase plane analysis.')

  def plot_vector_field(self, *args, **kwargs):
    """Plot vector filed of a 2D/1D system.

    Parameters
    ----------

    show : bool
        Whether show the figure.
    """
    self.analyzer.plot_vector_field(*args, **kwargs)

  def plot_fixed_point(self, *args, **kwargs):
    """Plot fixed points.

    Parameters
    ----------

    show : bool
        Whether show the figure.
    """
    return self.analyzer.plot_fixed_point(*args, **kwargs)

  def plot_nullcline(self, *args, **kwargs):
    """Plot nullcline (only supported in 2D system).

    Parameters
    ----------

    show : bool
        Whether show the figure.
    """
    self.analyzer.plot_nullcline(*args, **kwargs)

  def plot_trajectory(self, initials, duration, plot_duration=None, axes='v-v', show=False):
    """Plot trajectories according to the settings.

    Parameters
    ----------
    initials : list, tuple, dict
        The initial value setting of the targets. It can be a tuple/list of floats to specify
        each value of dynamical variables (for example, ``(a, b)``). It can also be a
        tuple/list of tuple to specify multiple initial values (for example,
        ``[(a1, b1), (a2, b2)]``).
    duration : int, float, tuple, list
        The running duration. Same with the ``duration`` in ``NeuGroup.run()``.
        It can be a int/float (``t_end``) to specify the same running end time,
        or it can be a tuple/list of int/float (``(t_start, t_end)``) to specify
        the start and end simulation time. Or, it can be a list of tuple
        (``[(t1_start, t1_end), (t2_start, t2_end)]``) to specify the specific
        start and end simulation time for each initial value.
    plot_duration : tuple, list, optional
        The duration to plot. It can be a tuple with ``(start, end)``. It can
        also be a list of tuple ``[(start1, end1), (start2, end2)]`` to specify
        the plot duration for each initial value running.
    axes : str
        The axes to plot. It can be:

             - 'v-v'
                    Plot the trajectory in the 'x_var'-'y_var' axis.
             - 't-v'
                    Plot the trajectory in the 'time'-'var' axis.
    show : bool
        Whether show or not.
    """
    self.analyzer.plot_trajectory(initials=initials,
                                  duration=duration,
                                  plot_duration=plot_duration,
                                  axes=axes,
                                  show=show)

  def plot_limit_cycle_by_sim(self, initials, duration, tol=0.001, show=False):
    """Plot limit cycles according to the settings.

    Parameters
    ----------
    initials : list, tuple
        The initial value setting of the targets. It can be a tuple/list of floats to specify
        each value of dynamical variables (for example, ``(a, b)``). It can also be a
        tuple/list of tuple to specify multiple initial values (for example,
        ``[(a1, b1), (a2, b2)]``).
    duration : int, float, tuple, list
        The running duration. Same with the ``duration`` in ``NeuGroup.run()``.
        It can be a int/float (``t_end``) to specify the same running end time,
        or it can be a tuple/list of int/float (``(t_start, t_end)``) to specify
        the start and end simulation time. Or, it can be a list of tuple
        (``[(t1_start, t1_end), (t2_start, t2_end)]``) to specify the specific
        start and end simulation time for each initial value.
    show : bool
        Whether show or not.
    """
    self.analyzer.plot_limit_cycle_by_sim(initials=initials,
                                          duration=duration,
                                          tol=tol,
                                          show=show)


class _PhasePlane1D(base.Base1DAnalyzer):
  """Phase plane analyzer for 1D system.
  """

  def plot_vector_field(self, show=False):
    """Plot the vector filed.

    Parameters
    ----------
    show : bool
        Whether show the figure.

    Returns
    -------
    results : np.ndarray
        The dx values.
    """
    logger.info('plot vector field ...')

    # 1. Nullcline of the x variable
    try:
      y_val = self.get_f_dx()(self.resolutions[self.x_var])
    except TypeError:
      raise errors.BrainPyError('Missing variables. Please check and set missing '
                                'variables to "fixed_vars".')

    # 2. visualization
    label = f"d{self.x_var}dt"
    x_style = dict(color='lightcoral', alpha=.7, linewidth=4)
    plt.plot(self.resolutions[self.x_var], y_val, **x_style, label=label)
    plt.axhline(0)

    plt.xlabel(self.x_var)
    plt.ylabel(label)
    plt.xlim(*utils.rescale(self.target_vars[self.x_var],
                            scale=(self.options.lim_scale - 1.) / 2))
    plt.legend()
    if show:
      plt.show()
    return y_val

  def plot_fixed_point(self, show=False):
    """Plot the fixed point.

    Parameters
    ----------
    show : bool
        Whether show the figure.

    Returns
    -------
    points : np.ndarray
        The fixed points.
    """
    logger.info('plot fixed point ...')

    # 1. functions
    f_fixed_point = self.get_f_fixed_point()
    f_dfdx = self.get_f_dfdx()

    # 2. stability analysis
    x_values = f_fixed_point()
    container = {a: [] for a in stability.get_1d_stability_types()}
    for i in range(len(x_values)):
      x = x_values[i]
      dfdx = f_dfdx(x)
      fp_type = stability.stability_analysis(dfdx)
      logger.info(f"Fixed point #{i + 1} at {self.x_var}={x} is a {fp_type}.")
      container[fp_type].append(x)

    # 3. visualization
    for fp_type, points in container.items():
      if len(points):
        plot_style = stability.plot_scheme[fp_type]
        plt.plot(points, [0] * len(points), '.',
                 markersize=20, **plot_style, label=fp_type)
    plt.legend()
    if show:
      plt.show()

    return np.array(x_values)

  def plot_nullcline(self, resolution=0.1, show=False):
    raise NotImplementedError('1D phase plane do not support plot_nullcline.')

  def plot_trajectory(self, *args, **kwargs):
    raise NotImplementedError('1D phase plane do not support plot_trajectory.')

  def plot_limit_cycle_by_sim(self, *args, **kwargs):
    raise NotImplementedError('1D phase plane do not support plot_limit_cycle_by_sim.')


class _PhasePlane2D(base.Base2DAnalyzer):
  """Phase plane analyzer for 2D system.
  """

  def plot_vector_field(self, plot_method='streamplot', plot_style=None, show=False):
    """Plot the vector field.

    Parameters
    ----------
    plot_method : str
        The method to plot the vector filed.
        It can be "streamplot" or "quiver".
    plot_style : dict, optional
        The style for vector filed plotting.

        - For ``plot_method="streamplot"``, it can set the keywords like "density",
          "linewidth", "color", "arrowsize". More settings please check
          https://matplotlib.org/api/_as_gen/matplotlib.pyplot.streamplot.html.
        - For ``plot_method="quiver"``, it can set the keywords like "color",
          "units", "angles", "scale". More settings please check
          https://matplotlib.org/api/_as_gen/matplotlib.pyplot.quiver.html.

    Returns
    -------
    result : tuple
        The ``dx``, ``dy`` values.
    """
    logger.info('plot vector field ...')

    if plot_style is None:
      plot_style = dict()

    xs = self.resolutions[self.x_var]
    ys = self.resolutions[self.y_var]
    X, Y = np.meshgrid(xs, ys)

    # dx
    try:
      dx = self.get_f_dx()(X, Y)
    except TypeError:
      raise errors.BrainPyError('Missing variables. Please check and set missing '
                                'variables to "fixed_vars".')

    # dy
    try:
      dy = self.get_f_dy()(X, Y)
    except TypeError:
      raise errors.BrainPyError('Missing variables. Please check and set missing '
                                'variables to "fixed_vars".')

    # vector field
    if plot_method == 'quiver':
      styles = dict()
      styles['units'] = plot_style.get('units', 'xy')
      if (not np.isnan(dx).any()) and (not np.isnan(dy).any()):
        speed = np.sqrt(dx ** 2 + dy ** 2)
        dx = dx / speed
        dy = dy / speed
      plt.quiver(X, Y, dx, dy, **styles)
    elif plot_method == 'streamplot':
      styles = dict()
      styles['arrowsize'] = plot_style.get('arrowsize', 1.2)
      styles['density'] = plot_style.get('density', 1)
      styles['color'] = plot_style.get('color', 'thistle')
      linewidth = plot_style.get('linewidth', None)
      if (linewidth is None) and (not np.isnan(dx).any()) and (not np.isnan(dy).any()):
        min_width = plot_style.get('min_width', 0.5)
        max_width = plot_style.get('min_width', 5.5)
        speed = np.sqrt(dx ** 2 + dy ** 2)
        linewidth = min_width + max_width * speed / speed.max()
      plt.streamplot(X, Y, dx, dy, linewidth=linewidth, **styles)
    else:
      raise ValueError(f'Unknown plot_method "{plot_method}", only supports "quiver" and "streamplot".')

    plt.xlabel(self.x_var)
    plt.ylabel(self.y_var)

    if show:
      plt.show()

    return dx, dy

  def plot_fixed_point(self, show=False):
    """Plot the fixed point and analyze its stability.

    Parameters
    ----------
    show : bool
        Whether show the figure.

    Returns
    -------
    results : tuple
        The value points.
    """
    logger.info('plot fixed point ...')

    # function for fixed point solving
    f_fixed_point = self.get_f_fixed_point()
    x_values, y_values = f_fixed_point()

    # function for jacobian matrix
    f_jacobian = self.get_f_jacobian()

    # stability analysis
    # ------------------
    container = {a: {'x': [], 'y': []} for a in stability.get_2d_stability_types()}
    for i in range(len(x_values)):
      x = x_values[i]
      y = y_values[i]
      fp_type = stability.stability_analysis(f_jacobian(x, y))
      logger.info(f"Fixed point #{i + 1} at {self.x_var}={x}, {self.y_var}={y} is a {fp_type}.")
      container[fp_type]['x'].append(x)
      container[fp_type]['y'].append(y)

    # visualization
    # -------------
    for fp_type, points in container.items():
      if len(points['x']):
        plot_style = stability.plot_scheme[fp_type]
        plt.plot(points['x'], points['y'], '.', markersize=20, **plot_style, label=fp_type)
    plt.legend()
    if show:
      plt.show()

    return x_values, y_values

  def plot_nullcline(self, numerical_setting=None, show=False):
    """Plot the nullcline.

    Parameters
    ----------
    numerical_setting : dict, optional
        Set the numerical method for solving nullclines.
        For each function setting, it contains the following keywords:

            coords
                The coordination setting, it can be 'var1-var2' (which means
                for each possible value 'var1' the optimizer method will search
                the zero root of 'var2') or 'var2-var1' (which means iterate each
                'var2' and get the optimization results of 'var1').
            plot
                It can be 'scatter' (default) or 'line'.

    show : bool
        Whether show the figure.

    Returns
    -------
    values : dict
        A dict with the format of ``{func1: (x_val, y_val), func2: (x_val, y_val)}``.
    """
    logger.info('plot nullcline ...')

    if numerical_setting is None:
      numerical_setting = dict()
    x_setting = numerical_setting.get(self.x_eq_group.func_name, {})
    y_setting = numerical_setting.get(self.y_eq_group.func_name, {})
    x_coords = x_setting.get('coords', self.x_var + '-' + self.y_var)
    y_coords = y_setting.get('coords', self.x_var + '-' + self.y_var)
    x_plot_style = x_setting.get('plot', 'scatter')
    y_plot_style = y_setting.get('plot', 'scatter')

    xs = self.resolutions[self.x_var]
    ys = self.resolutions[self.y_var]

    # Nullcline of the y variable
    y_style = dict(color='cornflowerblue', alpha=.7, )
    y_by_x = self.get_y_by_x_in_y_eq()
    if y_by_x['status'] == 'sympy_success':
      try:
        y_values_in_y_eq = y_by_x['f'](xs)
      except TypeError:
        raise errors.BrainPyError('Missing variables. Please check and set missing '
                                  'variables to "fixed_vars".')
      x_values_in_y_eq = xs
      plt.plot(xs, y_values_in_y_eq, **y_style, label=f"{self.y_var} nullcline")

    else:
      x_by_y = self.get_x_by_y_in_y_eq()
      if x_by_y['status'] == 'sympy_success':
        try:
          x_values_in_y_eq = x_by_y['f'](ys)
        except TypeError:
          raise errors.BrainPyError('Missing variables. Please check and set missing '
                                    'variables to "fixed_vars".')
        y_values_in_y_eq = ys
        plt.plot(x_values_in_y_eq, ys, **y_style, label=f"{self.y_var} nullcline")
      else:
        # optimization results
        optimizer = self.get_f_optimize_y_nullcline(y_coords)
        x_values_in_y_eq, y_values_in_y_eq = optimizer()

        if x_plot_style == 'scatter':
          plt.plot(x_values_in_y_eq, y_values_in_y_eq, '.', **y_style, label=f"{self.y_var} nullcline")
        elif x_plot_style == 'line':
          plt.plot(x_values_in_y_eq, y_values_in_y_eq, **y_style, label=f"{self.y_var} nullcline")
        else:
          raise ValueError(f'Unknown plot style: {x_plot_style}')

    # Nullcline of the x variable
    x_style = dict(color='lightcoral', alpha=.7, )
    y_by_x = self.get_y_by_x_in_x_eq()
    if y_by_x['status'] == 'sympy_success':
      try:
        y_values_in_x_eq = y_by_x['f'](xs)
      except TypeError:
        raise errors.BrainPyError('Missing variables. Please check and set missing '
                                  'variables to "fixed_vars".')
      x_values_in_x_eq = xs
      plt.plot(xs, y_values_in_x_eq, **x_style, label=f"{self.x_var} nullcline")

    else:
      x_by_y = self.get_x_by_y_in_x_eq()
      if x_by_y['status'] == 'sympy_success':
        try:
          x_values_in_x_eq = x_by_y['f'](ys)
        except TypeError:
          raise errors.BrainPyError('Missing variables. Please check and set missing '
                                    'variables to "fixed_vars".')
        y_values_in_x_eq = ys
        plt.plot(x_values_in_x_eq, ys, **x_style, label=f"{self.x_var} nullcline")
      else:
        # optimization results
        optimizer = self.get_f_optimize_x_nullcline(x_coords)
        x_values_in_x_eq, y_values_in_x_eq = optimizer()

        # visualization
        if y_plot_style == 'scatter':
          plt.plot(x_values_in_x_eq, y_values_in_x_eq, '.', **x_style, label=f"{self.x_var} nullcline")
        elif y_plot_style == 'line':
          plt.plot(x_values_in_x_eq, y_values_in_x_eq, **x_style, label=f"{self.x_var} nullcline")
        else:
          raise ValueError(f'Unknown plot style: {x_plot_style}')
    # finally
    plt.xlabel(self.x_var)
    plt.ylabel(self.y_var)
    scale = (self.options.lim_scale - 1.) / 2
    plt.xlim(*utils.rescale(self.target_vars[self.x_var], scale=scale))
    plt.ylim(*utils.rescale(self.target_vars[self.y_var], scale=scale))
    plt.legend()
    if show:
      plt.show()

    return {self.x_eq_group.func_name: (x_values_in_x_eq, y_values_in_x_eq),
            self.y_eq_group.func_name: (x_values_in_y_eq, y_values_in_y_eq)}

  def plot_trajectory(self, initials, duration, plot_duration=None, axes='v-v', show=False):
    """Plot trajectories according to the settings.

    Parameters
    ----------
    initials : list, tuple, dict
        The initial value setting of the targets. It can be a tuple/list of floats to specify
        each value of dynamical variables (for example, ``(a, b)``). It can also be a
        tuple/list of tuple to specify multiple initial values (for example,
        ``[(a1, b1), (a2, b2)]``).
    duration : int, float, tuple, list
        The running duration. Same with the ``duration`` in ``NeuGroup.run()``.
        It can be a int/float (``t_end``) to specify the same running end time,
        or it can be a tuple/list of int/float (``(t_start, t_end)``) to specify
        the start and end simulation time. Or, it can be a list of tuple
        (``[(t1_start, t1_end), (t2_start, t2_end)]``) to specify the specific
        start and end simulation time for each initial value.
    plot_duration : tuple, list, optional
        The duration to plot. It can be a tuple with ``(start, end)``. It can
        also be a list of tuple ``[(start1, end1), (start2, end2)]`` to specify
        the plot duration for each initial value running.
    axes : str
        The axes to plot. It can be:

             - 'v-v'
                    Plot the trajectory in the 'x_var'-'y_var' axis.
             - 't-v'
                    Plot the trajectory in the 'time'-'var' axis.
    show : bool
        Whether show or not.
    """

    logger.info('plot trajectory ...')

    if axes not in ['v-v', 't-v']:
      raise errors.BrainPyError(f'Unknown axes "{axes}", only support "v-v" and "t-v".')

    # 1. format the initial values
    if isinstance(initials, dict):
      initials = [initials]
    elif isinstance(initials, (list, tuple)):
      if isinstance(initials[0], (int, float)):
        initials = [{self.dvar_names[i]: v for i, v in enumerate(initials)}]
      elif isinstance(initials[0], dict):
        initials = initials
      elif isinstance(initials[0], (tuple, list)) and isinstance(initials[0][0], (int, float)):
        initials = [{self.dvar_names[i]: v for i, v in enumerate(init)} for init in initials]
      else:
        raise ValueError
    else:
      raise ValueError

    # 2. format the running duration
    if isinstance(duration, (int, float)):
      duration = [(0, duration) for _ in range(len(initials))]
    elif isinstance(duration[0], (int, float)):
      duration = [duration for _ in range(len(initials))]
    else:
      assert len(duration) == len(initials)

    # 3. format the plot duration
    if plot_duration is None:
      plot_duration = duration
    if isinstance(plot_duration[0], (int, float)):
      plot_duration = [plot_duration for _ in range(len(initials))]
    else:
      assert len(plot_duration) == len(initials)

    # 5. run the network
    for init_i, initial in enumerate(initials):
      traj_group = Trajectory(size=1,
                              integrals=self.model.integrals,
                              target_vars=initial,
                              fixed_vars=self.fixed_vars,
                              pars_update=self.pars_update,
                              scope=self.model.scopes)

      #   5.2 run the model
      traj_group.run(duration=duration[init_i], report=False, )

      #   5.3 legend
      legend = f'$traj_{init_i}$: '
      for key in self.dvar_names:
        legend += f'{key}={initial[key]}, '
      legend = legend[:-2]

      #   5.4 trajectory
      start = int(plot_duration[init_i][0] / math.get_dt())
      end = int(plot_duration[init_i][1] / math.get_dt())

      #   5.5 visualization
      if axes == 'v-v':
        lines = plt.plot(traj_group.mon[self.x_var][start: end, 0],
                         traj_group.mon[self.y_var][start: end, 0],
                         label=legend)
        utils.add_arrow(lines[0])
      else:
        plt.plot(traj_group.mon.ts[start: end],
                 traj_group.mon[self.x_var][start: end, 0],
                 label=legend + f', {self.x_var}')
        plt.plot(traj_group.mon.ts[start: end],
                 traj_group.mon[self.y_var][start: end, 0],
                 label=legend + f', {self.y_var}')

    # 6. visualization
    if axes == 'v-v':
      plt.xlabel(self.x_var)
      plt.ylabel(self.y_var)
      scale = (self.options.lim_scale - 1.) / 2
      plt.xlim(*utils.rescale(self.target_vars[self.x_var], scale=scale))
      plt.ylim(*utils.rescale(self.target_vars[self.y_var], scale=scale))
      plt.legend()
    else:
      plt.legend(title='Initial values')

    if show:
      plt.show()

  def plot_limit_cycle_by_sim(self, initials, duration, tol=0.001, show=False):
    """Plot trajectories according to the settings.

    Parameters
    ----------
    initials : list, tuple
        The initial value setting of the targets. It can be a tuple/list of floats to specify
        each value of dynamical variables (for example, ``(a, b)``). It can also be a
        tuple/list of tuple to specify multiple initial values (for example,
        ``[(a1, b1), (a2, b2)]``).
    duration : int, float, tuple, list
        The running duration. Same with the ``duration`` in ``NeuGroup.run()``.
        It can be a int/float (``t_end``) to specify the same running end time,
        or it can be a tuple/list of int/float (``(t_start, t_end)``) to specify
        the start and end simulation time. Or, it can be a list of tuple
        (``[(t1_start, t1_end), (t2_start, t2_end)]``) to specify the specific
        start and end simulation time for each initial value.
    show : bool
        Whether show or not.
    """
    logger.info('plot limit cycle ...')

    # 1. format the initial values
    if isinstance(initials, dict):
      initials = [initials]
    elif isinstance(initials, (list, tuple)):
      if isinstance(initials[0], (int, float)):
        initials = [{self.dvar_names[i]: v for i, v in enumerate(initials)}]
      elif isinstance(initials[0], dict):
        initials = initials
      elif isinstance(initials[0], (tuple, list)) and isinstance(initials[0][0], (int, float)):
        initials = [{self.dvar_names[i]: v for i, v in enumerate(init)} for init in initials]
      else:
        raise ValueError
    else:
      raise ValueError

    # 2. format the running duration
    if isinstance(duration, (int, float)):
      duration = [(0, duration) for _ in range(len(initials))]
    elif isinstance(duration[0], (int, float)):
      duration = [duration for _ in range(len(initials))]
    else:
      assert len(duration) == len(initials)

    # 5. run the network
    for init_i, initial in enumerate(initials):
      traj_group = Trajectory(size=1,
                              integrals=self.model.integrals,
                              target_vars=initial,
                              fixed_vars=self.fixed_vars,
                              pars_update=self.pars_update,
                              scope=self.model.scopes)

      #   5.2 run the model
      traj_group.run(duration=duration[init_i], report=False, )
      x_data = traj_group.mon[self.x_var][:, 0]
      y_data = traj_group.mon[self.y_var][:, 0]
      max_index = utils.find_indexes_of_limit_cycle_max(x_data, tol=tol)
      if max_index[0] != -1:
        x_cycle = x_data[max_index[0]: max_index[1]]
        y_cycle = y_data[max_index[0]: max_index[1]]
        # 5.5 visualization
        lines = plt.plot(x_cycle, y_cycle, label='limit cycle')
        utils.add_arrow(lines[0])
      else:
        logger.info(f'No limit cycle found for initial value {initial}')

    # 6. visualization
    plt.xlabel(self.x_var)
    plt.ylabel(self.y_var)
    scale = (self.options.lim_scale - 1.) / 2
    plt.xlim(*utils.rescale(self.target_vars[self.x_var], scale=scale))
    plt.ylim(*utils.rescale(self.target_vars[self.y_var], scale=scale))
    plt.legend()

    if show:
      plt.show()



# -*- coding: utf-8 -*-


__all__ = [
  'easy_scan',
  'easy_forloop',
  'easy_while',
  'easy_cond',
]


def easy_scan(f, dyn_vars, out_vars):
  raise ValueError


def easy_forloop(lower, upper, body_fun, init_val):
  raise ValueError


def easy_while(cond_fun, body_fun, init_val):
  raise ValueError


def easy_cond(pred, true_fun, false_fun, operand):
  raise ValueError



# -*- coding: utf-8 -*-

__all__ = [
  'size2num'
]


def size2num(size):
  if isinstance(size, int):
    return size
  elif isinstance(size, (tuple, list)):
    a = 1
    for b in size:
      a *= b
    return a
  else:
    raise ValueError



# -*- coding: utf-8 -*-

__version__ = "1.1.0rc6"


# "base" module
from . import base
from .base.base import *
from .base.collector import *


# "math" module
from . import math


# "integrators" module
from . import integrators
from .integrators import ode
from .integrators import sde
from .integrators.wrapper import *


# "simulation" module
from . import simulation
from .simulation.brainobjects import *
from .simulation.monitor import *
# submodules
from .simulation import brainobjects
from .simulation import connect
from .simulation import initialize
from .simulation import layers
from .simulation import nets
# py files
from .simulation import inputs
from .simulation import measure

# "analysis" module
from . import analysis
from .analysis import symbolic
from .analysis import continuation


# "visualization" module
from . import visualization as visualize


# other modules
from . import errors
from . import running
from . import tools




# -*- coding: utf-8 -*-


import brainpy as bp

bp.integrators.set_default_odeint('rk4')


class GABAa(bp.TwoEndConn):
  def __init__(self, pre, post, conn, delay=0., g_max=0.1, E=-75.,
               alpha=12., beta=0.1, T=1.0, T_duration=1.0, **kwargs):
    super(GABAa, self).__init__(pre=pre, post=post, conn=conn, **kwargs)

    # parameters
    self.g_max = g_max
    self.E = E
    self.alpha = alpha
    self.beta = beta
    self.T = T
    self.T_duration = T_duration
    self.delay = delay

    # connections
    self.conn_mat = self.conn.requires('conn_mat')
    self.size = bp.math.shape(self.conn_mat)

    # variables
    self.t_last_pre_spike = bp.math.ones(self.size) * -1e7
    self.s = bp.math.zeros(self.size)
    self.g = self.register_constant_delay('g', size=self.size, delay=delay)

  @staticmethod
  @bp.odeint
  def int_s(s, t, TT, alpha, beta):
    return alpha * TT * (1 - s) - beta * s

  def update(self, _t, _i, **kwargs):
    spike = bp.math.reshape(self.pre.spikes, (self.pre.num, 1)) * self.conn_mat
    self.t_last_pre_spike = bp.math.where(spike, _t, self.t_last_pre_spike)
    TT = ((_t - self.t_last_pre_spike) < self.T_duration) * self.T
    self.s = self.int_s(self.s, _t, TT, self.alpha, self.beta)
    self.g.push(self.g_max * self.s)
    g = self.g.pull()
    self.post.inputs -= bp.math.sum(g, axis=0) * (self.post.V - self.E)


class HH(bp.NeuGroup):
  def __init__(self, size, ENa=55., EK=-90., EL=-65,
               C=1.0, gNa=35., gK=9., gL=0.1, V_th=20.,
               phi=5.0, **kwargs):
    super(HH, self).__init__(size=size, **kwargs)

    # parameters
    self.ENa = ENa
    self.EK = EK
    self.EL = EL
    self.C = C
    self.gNa = gNa
    self.gK = gK
    self.gL = gL
    self.V_th = V_th
    self.phi = phi

    # variables
    self.V = bp.math.ones(self.num) * -65.
    self.h = bp.math.ones(self.num) * 0.6
    self.n = bp.math.ones(self.num) * 0.32
    self.spikes = bp.math.zeros(self.num)
    self.inputs = bp.math.zeros(self.num)

  @bp.odeint
  def integral(self, V, h, n, t, Iext):
    alpha = 0.07 * bp.math.exp(-(V + 58) / 20)
    beta = 1 / (bp.math.exp(-0.1 * (V + 28)) + 1)
    dhdt = alpha * (1 - h) - beta * h

    alpha = -0.01 * (V + 34) / (bp.math.exp(-0.1 * (V + 34)) - 1)
    beta = 0.125 * bp.math.exp(-(V + 44) / 80)
    dndt = alpha * (1 - n) - beta * n

    m_alpha = -0.1 * (V + 35) / (bp.math.exp(-0.1 * (V + 35)) - 1)
    m_beta = 4 * bp.math.exp(-(V + 60) / 18)
    m = m_alpha / (m_alpha + m_beta)
    INa = self.gNa * m ** 3 * h * (V - self.ENa)
    IK = self.gK * n ** 4 * (V - self.EK)
    IL = self.gL * (V - self.EL)
    dVdt = (- INa - IK - IL + Iext) / self.C

    return dVdt, self.phi * dhdt, self.phi * dndt

  def update(self, _t, _i, **kwargs):
    V, h, n = self.integral(self.V, self.h, self.n, _t, self.inputs)
    self.spikes[:] = (self.V < self.V_th) * (V >= self.V_th)
    self.V[:] = V
    self.h[:] = h
    self.n[:] = n
    self.inputs[:] = 0


def try1():
    num = 10
    neu = HH(num, monitors=['spikes', 'V'])
    neu.V = -70. + bp.math.random.normal(size=num) * 20

    syn = GABAa(pre=neu, post=neu, conn=bp.connect.All2All(include_self=False))
    syn.g_max = 0.1 / num

    net = bp.Network(neu=neu, syn=syn)

    for method in ['relative', 'absolute']:
      print(f'Method: {method}\n')
      print('vars:')
      print('-----')
      print('neu.vars()', list(neu.vars(method).keys()))
      print('syn.vars()', list(syn.vars(method).keys()))
      print('net.vars()', list(net.vars(method).keys()))
      print()

      print('ints:')
      print('-----')
      print('neu.ints()', list(neu.ints(method).keys()))
      print('syn.ints()', list(syn.ints(method).keys()))
      print('net.ints()', list(net.ints(method).keys()))
      print()

      print('nodes:')
      print('------')
      print('neu.nodes()', list(neu.nodes(method).keys()))
      print('syn.nodes()', list(syn.nodes(method).keys()))
      print('net.nodes()', list(net.nodes(method).keys()))
      print()


if __name__ == '__main__':
    try1()



# -*- coding: utf-8 -*-


from jax import lax

from brainpy.base import ArrayCollector

__all__ = [
  'easy_scan',
  'easy_forloop',
  'easy_while',
  'easy_cond',
]


def easy_scan(f, dyn_vars, out_vars):
  """Make a scan function.

  Parameters
  ----------
  f
  dyn_vars
  out_vars

  Returns
  -------

  """
  # dynamical variables
  if isinstance(dyn_vars, dict):
    dyn_vars = ArrayCollector(dyn_vars).unique()
  elif isinstance(dyn_vars, (tuple, list)):
    dyn_vars = ArrayCollector({f'_v{i}': v for i, v in enumerate(dyn_vars)}).unique()
  else:
    raise ValueError

  # outputs
  if isinstance(out_vars, dict):
    collector_outs = ArrayCollector(out_vars).unique()
    dict_return = True
  elif isinstance(out_vars, (tuple, list)):
    collector_outs = ArrayCollector({f'_o{i}': v for i, v in enumerate(out_vars)}).unique()
    dict_return = False
  else:
    raise ValueError

  # base function
  def fun2scan(init_dicts, x):
    dyn_vars.assign(init_dicts)
    f(x)
    return dyn_vars.dict(), collector_outs.dict()

  # return function
  if dict_return:
    def call(xs=None, length=None, reverse=False, unroll=1):
      init_dicts, all_outs = lax.scan(f=fun2scan, init=dyn_vars.dict(),
                                      xs=xs, length=length, reverse=reverse, unroll=unroll)
      dyn_vars.assign(init_dicts)
      return all_outs
  else:
    def call(xs=None, length=None, reverse=False, unroll=1):
      init_dicts, all_outs = lax.scan(f=fun2scan, init=dyn_vars.dict(),
                                      xs=xs, length=length, reverse=reverse, unroll=unroll)
      dyn_vars.assign(init_dicts)
      return tuple(all_outs.values())
  return call


def easy_forloop(lower, upper, body_fun, init_val):
  return lax.fori_loop(lower=lower, upper=upper, body_fun=body_fun, init_val=init_val)


def easy_while(cond_fun, body_fun, init_val):
  return lax.while_loop(cond_fun, body_fun, init_val)


def easy_cond(pred, true_fun, false_fun, operand):
  return lax.cond(pred, true_fun, false_fun, operand)



# -*- coding: utf-8 -*-

import pytest

from brainpy import errors
from brainpy.integrators import ode
from brainpy.integrators import odeint



def test_bs():
    method = 'bs'
    for adaptive in [True, False]:

        print(f'Test {"adaptive" if adaptive else ""} {method} method:')
        print()
        odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, p: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda f, t, dt: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, dt: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, v_new: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, dv_k1: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, dv_k2: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k2_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k2_t_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k3_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k3_t_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k4_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k4_t_arg: t)

        if adaptive:
            with pytest.raises(errors.CodeError):
                odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, error: t)

            with pytest.raises(errors.CodeError):
                odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, v_te: t)

            with pytest.raises(errors.CodeError):
                odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, dt_new: t)

        print('-' * 40)

def test_rkf45():
    method = 'rkf45'
    for adaptive in [True, False]:

        print(f'Test {"adaptive" if adaptive else ""} {method} method:')
        print()
        odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, p: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda f, t, dt: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, dt: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, v_new: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, dv_k1: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, dv_k2: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k2_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k2_t_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k3_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k3_t_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k4_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k4_t_arg: t)

        if adaptive:
            with pytest.raises(errors.CodeError):
                odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, error: t)

            with pytest.raises(errors.CodeError):
                odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, v_te: t)

            with pytest.raises(errors.CodeError):
                odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, dt_new: t)

        print('-' * 40)


def test_heun_euler():
    method = 'heun_euler'
    for adaptive in [True, False]:

        print(f'Test {"adaptive" if adaptive else ""} {method} method:')
        print()
        odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, p: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda f, t, dt: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, dt: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, v_new: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, dv_k1: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, dv_k2: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k2_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k2_t_arg: t)

        if adaptive:
            with pytest.raises(errors.CodeError):
                odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, error: t)

            with pytest.raises(errors.CodeError):
                odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, v_te: t)

            with pytest.raises(errors.CodeError):
                odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, dt_new: t)

        print('-' * 40)

def test_rkf12():
    method = 'rkf12'
    for adaptive in [True, False]:

        print(f'Test {"adaptive" if adaptive else ""} {method} method:')
        print()
        odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, p: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda f, t, dt: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, dt: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, v_new: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, dv_k1: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, dv_k2: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k2_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k2_t_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k3_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k3_t_arg: t)

        if adaptive:
            with pytest.raises(errors.CodeError):
                odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, error: t)

            with pytest.raises(errors.CodeError):
                odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, v_te: t)

            with pytest.raises(errors.CodeError):
                odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, dt_new: t)

        print('-' * 40)



def test_ck():
    method = 'ck'
    for adaptive in [True, False]:

        print(f'Test {"adaptive" if adaptive else ""} {method} method:')
        print()
        odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, p: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda f, t, dt: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, dt: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, v_new: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, dv_k1: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, dv_k2: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k2_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k2_t_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k3_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k3_t_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k4_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k4_t_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k5_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k5_t_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k6_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k6_t_arg: t)

        if adaptive:
            with pytest.raises(errors.CodeError):
                odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, error: t)

            with pytest.raises(errors.CodeError):
                odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, v_te: t)

            with pytest.raises(errors.CodeError):
                odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, dt_new: t)

        print('-' * 40)




def test_rkdp():
    method = 'rkdp'
    for adaptive in [True, False]:

        print(f'Test {"adaptive" if adaptive else ""} {method} method:')
        print()
        odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, p: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda f, t, dt: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, dt: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, v_new: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, dv_k1: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, dv_k2: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k2_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k2_t_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k3_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k3_t_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k4_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k4_t_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k5_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k5_t_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k6_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k6_t_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k7_v_arg: t)

        with pytest.raises(errors.CodeError):
            odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, k7_t_arg: t)

        if adaptive:
            with pytest.raises(errors.CodeError):
                odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, error: t)

            with pytest.raises(errors.CodeError):
                odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, v_te: t)

            with pytest.raises(errors.CodeError):
                odeint(method=method, show_code=True, adaptive=adaptive, f=lambda v, t, dt_new: t)

        print('-' * 40)





if __name__ == '__main__':
    test_bs()
    test_heun_euler()
    test_ck()
    test_rkdp()
    test_rkf12()
    test_rkf45()



# -*- coding: utf-8 -*-

from brainpy import errors, math
from brainpy.base.collector import Collector
from brainpy.simulation import utils
from brainpy.simulation.brainobjects.base import DynamicalSystem

__all__ = [
  'NeuGroup',
  'Channel',
  'CondNeuGroup',
  'Soma',
  'Dendrite',
]


class NeuGroup(DynamicalSystem):
  """Base class to model neuronal groups.

  There are several essential attributes:

  - ``size``: the geometry of the neuron group. For example, `(10, )` denotes a line of
    neurons, `(10, 10)` denotes a neuron group aligned in a 2D space, `(10, 15, 4)` denotes
    a 3-dimensional neuron group.
  - ``num``: the flattened number of neurons in the group. For example, `size=(10, )` => \
    `num=10`, `size=(10, 10)` => `num=100`, `size=(10, 15, 4)` => `num=600`.
  - ``shape``: the variable shape with `(num, num_batch)`.

  Parameters
  ----------
  size : int, tuple of int, list of int
    The neuron group geometry.
  num_batch : optional, int
    The batch size.
  steps : tuple of str, tuple of function, dict of (str, function), optional
    The step functions.
  steps : tuple of str, tuple of function, dict of (str, function), optional
    The callable function, or a list of callable functions.
  monitors : None, list, tuple, datastructures.Monitor
    Variables to monitor.
  name : optional, str
    The name of the dynamic system.
  """

  def __init__(self, size, num_batch=None, name=None, steps=('update',), **kwargs):
    # size
    if isinstance(size, (list, tuple)):
      if len(size) <= 0:
        raise errors.BrainPyError('size must be int, or a tuple/list of int.')
      if not isinstance(size[0], int):
        raise errors.BrainPyError('size must be int, or a tuple/list of int.')
      size = tuple(size)
    elif isinstance(size, int):
      size = (size,)
    else:
      raise errors.BrainPyError('size must be int, or a tuple/list of int.')
    self.size = size
    self.num = utils.size2len(size)
    self.num_batch = num_batch
    if num_batch is None:
      self.shape = (self.num, )
    else:
      self.shape = (self.num, self.num_batch)
      self.num_batch = num_batch
      assert isinstance(num_batch, int), 'Only support int for "num_batch"'

    # initialize
    super(NeuGroup, self).__init__(steps=steps, name=name, **kwargs)

  def update(self, _t, _dt, **kwargs):
    """The function to specify the updating rule.

    Parameters
    ----------
    _t : float
      The current time.
    _dt : float
      The time step.
    """
    raise NotImplementedError(f'Subclass of {self.__class__.__name__} must '
                              f'implement "update" function.')


# ----------------------------------------------------
#
#         Conductance-based Neuron Model
#
# ----------------------------------------------------


class Channel(DynamicalSystem):
  """Base class to model ion channels.

  Notes
  -----

  The ``__init__()`` function in :py:class:`Channel` is used to specify
  the parameters of the channel. The ``__call__()`` function
  is used to initialize the variables in this channel.
  """

  def __init__(self, **kwargs):
    super(Channel, self).__init__(**kwargs)

  def init(self, host, **kwargs):
    """Initialize variables in this channel."""
    if not isinstance(host, CondNeuGroup):
      raise errors.BrainPyError(f'Only support host with instance of {str(DynamicalSystem)}, while we got {host}')
    self.host = host

  def update(self, _t, _dt, **kwargs):
    """The function to specify the updating rule."""
    raise NotImplementedError(f'Subclass of {self.__class__.__name__} '
                              f'must implement "update" function.')


class CondNeuGroup(NeuGroup):
  """Base class to model conductance-based neuron group.

  The standard formulation for a conductance-based model is given as

  .. math::

      C_m {dV \over dt} = \sum_jg_j(E - V) + I_{ext}

  where :math:`g_j=\bar{g}_{j} M^x N^y` is the channel conductance, :math:`E` is the
  reversal potential, :math:`M` is the activation variable, and :math:`N` is the
  inactivation variable.

  :math:`M` and :math:`N` have the dynamics of

  .. math::

      {dx \over dt} = \phi_x {x_\infty (V) - x \over \tau_x(V)}

  where :math:`x \in [M, N]`, :math:`\phi_x` is a temperature-dependent factor,
  :math:`x_\infty` is the steady state, and :math:`\tau_x` is the time constant.
  Equivalently, the above equation can be written as:

  .. math::

      \frac{d x}{d t}=\phi_{x}\left(\alpha_{x}(1-x)-\beta_{x} x\right)

  where :math:`\alpha_{x}` and :math:`\beta_{x}` are rate constants.

  Parameters
  ----------
  size : int, tuple of int
    The network size of this neuron group.
  num_batch : optional, int
    The batch size.
  monitors : optional, list of str, tuple of str
    The neuron group monitor.
  name : optional, str
    The neuron group name.

  Notes
  -----

  The ``__init__()`` function in :py:class:`CondNeuGroup` is used to specify
  the parameters of channels and this neuron group. The ``__call__()`` function
  is used to initialize the variables in this neuron group.
  """

  def __init__(self, C=1., A=1e-3, V_th=0., **channels):
    # parameters for neuron
    self.C = C
    self.A = A
    self.V_th = V_th

    # children channels
    self.child_channels = Collector()
    for key, ch in channels.items():
      if not isinstance(ch, Channel):
        raise errors.BrainPyError(f'{self.__class__.__name__} only receives {Channel.__name__} '
                                  f'instance, while we got {type(ch)}: {ch}.')
      self.child_channels[key] = ch

  def init(self, size, num_batch=None, monitors=None, steps=('update',), name=None):
    super(CondNeuGroup, self).__init__(size, num_batch=num_batch, steps=steps, monitors=monitors, name=name)

    # initialize variables
    self.V = math.Variable(math.zeros(self.shape, dtype=math.float_))
    self.spike = math.Variable(math.zeros(self.shape, dtype=math.bool_))
    self.input = math.Variable(math.zeros(self.shape, dtype=math.float_))
    self.I_ion = math.Variable(math.zeros(self.shape, dtype=math.float_))
    self.V_linear = math.Variable(math.zeros(self.shape, dtype=math.float_))

    # initialize variables in channels
    for ch in self.child_channels.values(): ch.init(host=self)
    return self

  def update(self, _t, _dt, **kwargs):
    # update variables in channels
    for ch in self.child_channels.values():
      ch.update(_t, _dt)

    # update V using Exponential Euler method
    dvdt = self.I_ion / self.C + self.input * (1e-3 / self.A / self.C)
    linear = self.V_linear / self.C
    V = self.V + dvdt * (math.exp(linear * _dt) - 1) / linear

    # update other variables
    self.spike[:] = math.logical_and(V >= self.V_th, self.V < self.V_th)
    self.V_linear[:] = 0.
    self.I_ion[:] = 0.
    self.input[:] = 0.
    self.V[:] = V

  def __getattr__(self, item):
    """Wrap the dot access ('self.'). """
    child_channels = super(CondNeuGroup, self).__getattribute__('child_channels')
    if item in child_channels:
      return child_channels[item]
    else:
      return super(CondNeuGroup, self).__getattribute__(item)


# ---------------------------------------------------------
#
#          Multi-Compartment Neuron Model
#
# ---------------------------------------------------------

class Dendrite(DynamicalSystem):
  """Base class to model dendrites.

  """

  def __init__(self, name, **kwargs):
    super(Dendrite, self).__init__(name=name, **kwargs)


class Soma(DynamicalSystem):
  """Base class to model soma in multi-compartment neuron models.

  """

  def __init__(self, name, **kwargs):
    super(Soma, self).__init__(name=name, **kwargs)



# -*- coding: utf-8 -*-

import pytest

import brainpy as bp
from brainpy.integrators.ode.wrapper import exp_euler_wrapper


def test1():
  import numpy as np

  def drivative(V, m, h, n, t, Iext, gNa, ENa, gK, EK, gL, EL, C):
    alpha = 0.1 * (V + 40) / (1 - np.exp(-(V + 40) / 10))
    beta = 4.0 * np.exp(-(V + 65) / 18)
    dmdt = alpha * (1 - m) - beta * m

    alpha = 0.07 * np.exp(-(V + 65) / 20.)
    beta = 1 / (1 + np.exp(-(V + 35) / 10))
    dhdt = alpha * (1 - h) - beta * h

    alpha = 0.01 * (V + 55) / (1 - np.exp(-(V + 55) / 10))
    beta = 0.125 * np.exp(-(V + 65) / 80)
    dndt = alpha * (1 - n) - beta * n

    I_Na = (gNa * m ** 3.0 * h) * (V - ENa)
    I_K = (gK * n ** 4.0) * (V - EK)
    I_leak = gL * (V - EL)
    dVdt = (- I_Na - I_K - I_leak + Iext) / C

    return dVdt, dmdt, dhdt, dndt

  exp_euler_wrapper(f=drivative, show_code=True, dt=0.01, var_type='SCALAR', )


def test2():
  def derivative(s, t, tau):
    return -s / tau

  with pytest.raises(bp.errors.DiffEqError):
    exp_euler_wrapper(f=derivative, show_code=True, dt=0.01, var_type='SCALAR', )


def test3():
  def derivative(s, v, t, tau):
    dv = -v + 1
    return -s / tau, dv

  with pytest.raises(bp.errors.DiffEqError):
    exp_euler_wrapper(f=derivative, show_code=True, dt=0.01, var_type='SCALAR', )


def test4():
  f = lambda s, t, tau: -s / tau
  with pytest.raises(bp.errors.AnalyzerError) as excinfo:
    exp_euler_wrapper(f=f, show_code=True, dt=0.01, var_type='SCALAR', )


def test_derivative_with_self():
  def drivative(self, V, m, h, n, t, Iext):
    alpha = 0.1 * (V + 40) / (1 - bp.math.exp(-(V + 40) / 10))
    beta = 4.0 * bp.math.exp(-(V + 65) / 18)
    dmdt = alpha * (1 - m) - beta * m

    alpha = 0.07 * bp.math.exp(-(V + 65) / 20.)
    beta = 1 / (1 + bp.math.exp(-(V + 35) / 10))
    dhdt = alpha * (1 - h) - beta * h

    alpha = 0.01 * (V + 55) / (1 - bp.math.exp(-(V + 55) / 10))
    beta = 0.125 * bp.math.exp(-(V + 65) / 80)
    dndt = alpha * (1 - n) - beta * n

    I_Na = (self.gNa * m ** 3.0 * h) * (V - self.ENa)
    I_K = (self.gK * n ** 4.0) * (V - self.EK)
    I_leak = self.gL * (V - self.EL)
    dVdt = (- I_Na - I_K - I_leak + Iext) / self.C

    return dVdt, dmdt, dhdt, dndt

  print()
  exp_euler_wrapper(f=drivative, show_code=True, dt=0.01, var_type='SCALAR', )
